from numba import jit
import numpy as np
import math


def _standard_grid_numpy_wrap(vis_data, uvw, weight, freq_chan, cgk_1D, grid_parms):
    """
    Wraps the jit gridder code.

    Parameters
    ----------
    grid : complex array
        (n_chan, n_pol, n_u, n_v)
    sum_weight : float array
        (n_chan, n_pol)
    uvw  : float array
        (n_time, n_baseline, 3)
    freq_chan : float array
        (n_chan)
    weight : float array
        (n_time, n_baseline, n_vis_chan)
    cgk_1D : float array
        (oversampling*(support//2 + 1))
    grid_parms : dictionary
        keys ('image_size','cell','oversampling','support')

    Returns
    -------
    grid : complex array
        (1,n_imag_chan,n_imag_pol,n_u,n_v)
    """

    n_chan = weight.shape[2]
    if grid_parms["chan_mode"] == "cube":
        n_imag_chan = n_chan
        chan_map = (np.arange(0, n_chan)).astype(int)
    else:  # continuum
        n_imag_chan = 1  # Making only one continuum image.
        chan_map = (np.zeros(n_chan)).astype(int)

    n_imag_pol = weight.shape[3]
    pol_map = (np.arange(0, n_imag_pol)).astype(int)

    n_uv = grid_parms["image_size_padded"]
    delta_lm = grid_parms["cell_size"]
    oversampling = grid_parms["oversampling"]
    support = grid_parms["support"]

    if grid_parms["complex_grid"]:
        grid = np.zeros(
            (n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.complex128
        )
    else:
        grid = np.zeros((n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.double)
    sum_weight = np.zeros((n_imag_chan, n_imag_pol), dtype=np.double)

    do_psf = grid_parms["do_psf"]
    do_imaging_weight = False
    _standard_grid_jit(
        grid,
        sum_weight,
        do_psf,
        do_imaging_weight,
        vis_data,
        uvw,
        freq_chan,
        chan_map,
        pol_map,
        weight,
        cgk_1D,
        n_uv,
        delta_lm,
        support,
        oversampling,
    )

    return grid, sum_weight


def _standard_grid_psf_numpy_wrap(uvw, weight, freq_chan, cgk_1D, grid_parms):
    """
    Wraps the jit gridder code.

    Parameters
    ----------
    grid : complex array
        (n_chan, n_pol, n_u, n_v)
    sum_weight : float array
        (n_chan, n_pol)
    uvw  : float array
        (n_time, n_baseline, 3)
    freq_chan : float array
        (n_chan)
    weight : float array
        (n_time, n_baseline, n_vis_chan)
    cgk_1D : float array
        (oversampling*(support//2 + 1))
    grid_parms : dictionary
        keys ('image_size','cell','oversampling','support')

    Returns
    -------
    grid : complex array
        (1,n_imag_chan,n_imag_pol,n_u,n_v)
    """

    n_chan = weight.shape[2]
    if grid_parms["chan_mode"] == "cube":
        n_imag_chan = n_chan
        chan_map = (np.arange(0, n_chan)).astype(int)
    else:  # continuum
        n_imag_chan = 1  # Making only one continuum image.
        chan_map = (np.zeros(n_chan)).astype(int)

    n_imag_pol = weight.shape[3]
    pol_map = (np.arange(0, n_imag_pol)).astype(int)

    n_uv = grid_parms["image_size_padded"]
    delta_lm = grid_parms["cell_size"]
    oversampling = grid_parms["oversampling"]
    support = grid_parms["support"]

    grid = np.zeros((n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.double)
    sum_weight = np.zeros((n_imag_chan, n_imag_pol), dtype=np.double)

    do_imaging_weight = grid_parms["do_imaging_weight"]

    do_psf = grid_parms["do_psf"]
    vis_data = np.zeros(
        (1, 1, 1, 1), dtype=bool
    )  # This 0 bool array is needed to pass to _standard_grid_jit so that the code can be resued and to keep numba happy.
    # print('do_imaging_weight',do_imaging_weight)
    _standard_grid_jit(
        grid,
        sum_weight,
        do_psf,
        do_imaging_weight,
        vis_data,
        uvw,
        freq_chan,
        chan_map,
        pol_map,
        weight,
        cgk_1D,
        n_uv,
        delta_lm,
        support,
        oversampling,
    )
    # print('new code')
    return grid, sum_weight


import numpy as np


# When jit is used round is repolaced by standard c++ round that is different to python round
@jit(nopython=True, cache=True, nogil=True)  # fastmath=True
def _standard_grid_jit(
    grid,
    sum_weight,
    do_psf,
    do_imaging_weight,
    vis_data,
    uvw,
    freq_chan,
    chan_map,
    pol_map,
    weight,
    cgk_1D,
    n_uv,
    delta_lm,
    support,
    oversampling,
):
    """
    Parameters
    ----------
    grid : complex array
        (n_chan, n_pol, n_u, n_v)
    sum_weight : float array
        (n_chan, n_pol)
    vis_data : complex array
        (n_time, n_baseline, n_vis_chan, n_pol)
    uvw  : float array
        (n_time, n_baseline, 3)
    freq_chan : float array
        (n_chan)
    chan_map : int array
        (n_chan)
    pol_map : int array
        (n_pol)
    weight : float array
        (n_time, n_baseline, n_vis_chan)
    cgk_1D : float array
        (oversampling*(support//2 + 1))
    grid_parms : dictionary
        keys ('n_imag_chan','n_imag_pol','n_uv','delta_lm','oversampling','support')

    Returns
    -------
    """

    # By hardcoding the support and oversampling values, the innermost for loops can be unrolled by the compiler leading to significantly faster code.
    # support = 7
    # oversampling = 100

    c = 299792458.0
    uv_scale = np.zeros((2, len(freq_chan)), dtype=np.double)
    uv_scale[0, :] = -(freq_chan * delta_lm[0] * n_uv[0]) / c
    uv_scale[1, :] = -(freq_chan * delta_lm[1] * n_uv[1]) / c

    # oversampling_center = int(oversampling // 2)
    support_center = int(support // 2)
    uv_center = n_uv // 2

    start_support = -support_center
    end_support = (
        support - support_center
    )  # end_support is larger by 1 so that python range() gives correct indices

    n_time = uvw.shape[0]
    n_baseline = uvw.shape[1]
    n_chan = len(chan_map)
    n_pol = len(pol_map)

    n_u = n_uv[0]
    n_v = n_uv[1]

    for i_time in range(n_time):
        for i_baseline in range(n_baseline):
            for i_chan in range(n_chan):
                a_chan = chan_map[i_chan]
                u = uvw[i_time, i_baseline, 0] * uv_scale[0, i_chan]
                v = uvw[i_time, i_baseline, 1] * uv_scale[1, i_chan]

                if ~np.isnan(u) and ~np.isnan(v):
                    u_pos = u + uv_center[0]
                    v_pos = v + uv_center[1]

                    u_pos_conj = -u + uv_center[0]
                    v_pos_conj = -v + uv_center[1]

                    # Doing round as int(x+0.5) since u_pos/v_pos should always positive and this matices fortran and gives consistant rounding.
                    # u_center_indx = int(u_pos + 0.5)
                    # v_center_indx = int(v_pos + 0.5)

                    # Do not use numpy round
                    u_center_indx = int(u_pos + 0.5)
                    v_center_indx = int(v_pos + 0.5)

                    u_center_indx_conj = int(u_pos_conj + 0.5)
                    v_center_indx_conj = int(v_pos_conj + 0.5)

                    if (
                        (u_center_indx + support_center < n_u)
                        and (v_center_indx + support_center < n_v)
                        and (u_center_indx - support_center >= 0)
                        and (v_center_indx - support_center >= 0)
                    ):
                        u_offset = u_center_indx - u_pos
                        u_center_offset_indx = math.floor(u_offset * oversampling + 0.5)
                        v_offset = v_center_indx - v_pos
                        v_center_offset_indx = math.floor(v_offset * oversampling + 0.5)

                        for i_pol in range(n_pol):
                            if do_psf:
                                if (n_pol >= 2) and do_imaging_weight:
                                    weighted_data = (
                                        weight[i_time, i_baseline, i_chan, 0]
                                        + weight[i_time, i_baseline, i_chan, 1]
                                    ) / 2.0
                                    sel_weight = weighted_data
                                else:
                                    sel_weight = weight[
                                        i_time, i_baseline, i_chan, i_pol
                                    ]
                                    weighted_data = weight[
                                        i_time, i_baseline, i_chan, i_pol
                                    ]
                            else:
                                sel_weight = weight[i_time, i_baseline, i_chan, i_pol]
                                weighted_data = (
                                    vis_data[i_time, i_baseline, i_chan, i_pol]
                                    * weight[i_time, i_baseline, i_chan, i_pol]
                                )

                            # print('1. u_center_indx, v_center_indx', u_center_indx, v_center_indx, vis_data[i_time, i_baseline, i_chan, i_pol], weight[i_time, i_baseline, i_chan, i_pol])

                            if ~np.isnan(weighted_data) and (weighted_data != 0.0):
                                a_pol = pol_map[i_pol]
                                norm = 0.0

                                for i_v in range(start_support, end_support):
                                    v_indx = v_center_indx + i_v
                                    v_offset_indx = np.abs(
                                        oversampling * i_v + v_center_offset_indx
                                    )
                                    conv_v = cgk_1D[v_offset_indx]

                                    if do_imaging_weight:
                                        v_indx_conj = v_center_indx_conj + i_v

                                    for i_u in range(start_support, end_support):
                                        u_indx = u_center_indx + i_u
                                        u_offset_indx = np.abs(
                                            oversampling * i_u + u_center_offset_indx
                                        )
                                        conv_u = cgk_1D[u_offset_indx]
                                        conv = conv_u * conv_v

                                        grid[a_chan, a_pol, u_indx, v_indx] = (
                                            grid[a_chan, a_pol, u_indx, v_indx]
                                            + conv * weighted_data
                                        )
                                        norm = norm + conv

                                        if do_imaging_weight:
                                            u_indx_conj = u_center_indx_conj + i_u
                                            grid[
                                                a_chan, a_pol, u_indx_conj, v_indx_conj
                                            ] = (
                                                grid[
                                                    a_chan,
                                                    a_pol,
                                                    u_indx_conj,
                                                    v_indx_conj,
                                                ]
                                                + conv * weighted_data
                                            )

                                sum_weight[a_chan, a_pol] = (
                                    sum_weight[a_chan, a_pol] + sel_weight * norm
                                )

                                if do_imaging_weight:
                                    sum_weight[a_chan, a_pol] = (
                                        sum_weight[a_chan, a_pol] + sel_weight * norm
                                    )

    return


# TransformMachines/FTMachine.tcc
# template <class T> void FTMachine::getGrid(casacore::Array<T>& thegrid)
############################################################################################################################################################################################################################################################################################################################################################################################################################################################
############################################################################################################################################################################################################################################################################################################################################################################################################################################################


def _standard_imaging_weight_degrid_numpy_wrap(
    grid_imaging_weight,
    uvw,
    natural_imaging_weight,
    briggs_factors,
    freq_chan,
    grid_parms,
):
    n_chan = natural_imaging_weight.shape[2]
    n_imag_chan = n_chan

    #    if grid_parms['chan_mode'] == 'cube':
    #        n_imag_chan = n_chan
    #        chan_map = (np.arange(0, n_chan)).astype(int)
    #    else:  # continuum
    #        n_imag_chan = 1
    #        chan_map = (np.zeros(n_chan)).astype(int)
    # Should always be cube for image weights
    n_imag_chan = n_chan
    chan_map = (np.arange(0, n_chan)).astype(int)

    n_imag_pol = natural_imaging_weight.shape[3]
    pol_map = (np.arange(0, n_imag_pol)).astype(int)

    n_uv = grid_parms["image_size_padded"]
    delta_lm = grid_parms["cell_size"]

    imaging_weight = np.zeros(natural_imaging_weight.shape, dtype=np.double)

    _standard_imaging_weight_degrid_jit(
        imaging_weight,
        grid_imaging_weight,
        briggs_factors,
        uvw,
        freq_chan,
        chan_map,
        pol_map,
        natural_imaging_weight,
        n_uv,
        delta_lm,
    )

    return imaging_weight


@jit(nopython=True, cache=True, nogil=True)
def _standard_imaging_weight_degrid_jit(
    imaging_weight,
    grid_imaging_weight,
    briggs_factors,
    uvw,
    freq_chan,
    chan_map,
    pol_map,
    natural_imaging_weight,
    n_uv,
    delta_lm,
):
    c = 299792458.0
    uv_scale = np.zeros((2, len(freq_chan)), dtype=np.double)
    uv_scale[0, :] = -(freq_chan * delta_lm[0] * n_uv[0]) / c
    uv_scale[1, :] = -(freq_chan * delta_lm[1] * n_uv[1]) / c

    uv_center = n_uv // 2

    n_time = uvw.shape[0]
    n_baseline = uvw.shape[1]
    n_chan = len(chan_map)
    n_pol = len(pol_map)
    n_imag_chan = chan_map.shape[0]

    n_u = n_uv[0]
    n_v = n_uv[1]

    # print('Degrid operation')

    for i_time in range(n_time):
        for i_baseline in range(n_baseline):
            for i_chan in range(n_chan):
                a_chan = chan_map[i_chan]
                u = uvw[i_time, i_baseline, 0] * uv_scale[0, i_chan]
                v = uvw[i_time, i_baseline, 1] * uv_scale[1, i_chan]
                if ~np.isnan(u) and ~np.isnan(v):
                    u_pos = u + uv_center[0]
                    v_pos = v + uv_center[1]

                    # Doing round as int(x+0.5) since u_pos/v_pos should always be positive and  fortran and gives consistant rounding.
                    u_center_indx = int(u_pos + 0.5)
                    v_center_indx = int(v_pos + 0.5)

                    # print('f uv', freq_chan[i_chan], uvw[i_time, i_baseline, 0],uvw[i_time, i_baseline, 1])
                    if (
                        (u_center_indx < n_u)
                        and (v_center_indx < n_v)
                        and (u_center_indx >= 0)
                        and (v_center_indx >= 0)
                    ):
                        # print('u_center_indx, v_center_indx',  u_center_indx, v_center_indx)

                        for i_pol in range(n_pol):
                            a_pol = pol_map[i_pol]

                            if n_pol == 2:
                                imaging_weight[i_time, i_baseline, i_chan, i_pol] = (
                                    natural_imaging_weight[
                                        i_time, i_baseline, i_chan, 0
                                    ]
                                    + natural_imaging_weight[
                                        i_time, i_baseline, i_chan, 1
                                    ]
                                ) / 2.0
                            else:
                                imaging_weight[i_time, i_baseline, i_chan, i_pol] = (
                                    natural_imaging_weight[
                                        i_time, i_baseline, i_chan, i_pol
                                    ]
                                )

                            if ~np.isnan(
                                natural_imaging_weight[
                                    i_time, i_baseline, i_chan, i_pol
                                ]
                            ) and (
                                natural_imaging_weight[
                                    i_time, i_baseline, i_chan, i_pol
                                ]
                                != 0.0
                            ):
                                if ~np.isnan(
                                    grid_imaging_weight[
                                        a_chan, a_pol, u_center_indx, v_center_indx
                                    ]
                                ) and (
                                    grid_imaging_weight[
                                        a_chan, a_pol, u_center_indx, v_center_indx
                                    ]
                                    != 0.0
                                ):
                                    briggs_grid_imaging_weight = (
                                        briggs_factors[0, a_chan, a_pol]
                                        * grid_imaging_weight[
                                            a_chan, a_pol, u_center_indx, v_center_indx
                                        ]
                                        + briggs_factors[1, a_chan, a_pol]
                                    )
                                    imaging_weight[
                                        i_time, i_baseline, i_chan, i_pol
                                    ] = (
                                        imaging_weight[
                                            i_time, i_baseline, i_chan, i_pol
                                        ]
                                        / briggs_grid_imaging_weight
                                    )

    return



from numcodecs import Blosc


def cube_imaging_niter0(
    ps_name,
    image_name,
    grid_params,
    polarization_coord=None,
    frequency_coord=None,
    n_chunks=None,
    data_variables=["sky", "point_spread_function", "primary_beam"],
    intents=["OBSERVE_TARGET#ON_SOURCE"],
    compressor=Blosc(cname="lz4", clevel=5),
    data_group="base",
    double_precission=True,
    thread_info=None,
    workflow_ochestrator: {"dask", "airflow"} = "dask",
    airflow_dag_folder=None,
):
    import numpy as np
    import xarray as xr
    import dask
    import os
    from xradio.measurement_set import open_processing_set
    from graphviper.graph_tools.coordinate_utils import make_parallel_coord
    from graphviper.graph_tools import generate_dask_workflow, generate_airflow_workflow
    from graphviper.graph_tools import map, reduce
    from astroviper.imaging._utils import _make_image
    from xradio.image import make_empty_sky_image
    from xradio.image import write_image
    import zarr
    import toolviper.utils.logger as logger

    # Get metadata
    ps = open_processing_set(ps_name, intents=intents)
    summary_df = ps.summary()

    # Get phase center of mosaic if field given.
    if isinstance(grid_params["phase_direction"], int):
        ms_xds_name = summary_df.loc[
            summary_df["field_name"].index.get_loc(grid_params["phase_direction"])
        ]["name"]
        vis_name = ps[ms_xds_name].attrs["data_groups"][data_group]["correlated_data"]
        grid_params["phase_direction"] = ps[ms_xds_name][vis_name].field_and_source_xds[
            "FIELD_PHASE_CENTER"
        ]

    # Create Cube frequency axis
    if frequency_coord is None:
        frequency_coord = ps.get_ps_freq_axis()

    if polarization_coord is None:
        polarization_coord = ps.get(0).polarization

    # Create empty image
    img_xds = make_empty_sky_image(
        phase_center=grid_params["phase_direction"].values,
        image_size=grid_params["image_size"],
        cell_size=grid_params["cell_size"],
        chan_coords=frequency_coord.data,
        pol_coords=polarization_coord.data,
        time_coords=[0],
    )
    write_image(img_xds, imagename=image_name, out_format="zarr")

    if n_chunks is None:
        # Calculate n_chunks
        from astroviper._utils.data_partitioning import bytes_in_dtype

        ## Determine the amount of memory required by the node task if all dimensions that chunking will occur on are singleton.
        ## For example cube_imaging does chunking only only frequency, so memory_singleton_chunk should be the amount of memory requered by _feather when there is a single frequency channel.

        n_pixels_single_frequency = (
            img_xds.sizes["l"]
            * img_xds.sizes["m"]
            * img_xds.sizes["polarization"]
            * img_xds.sizes["time"]
        )
        fudge_factor = 1.2
        if double_precission:
            memory_singleton_chunk = fudge_factor * (
                3 * n_pixels_single_frequency * bytes_in_dtype["complex64"] / (1024**3)
                + 3 * n_pixels_single_frequency * bytes_in_dtype["float32"] / (1024**3)
            )
        else:
            memory_singleton_chunk = fudge_factor * (
                3 * n_pixels_single_frequency * bytes_in_dtype["complex128"] / (1024**3)
                + 3 * n_pixels_single_frequency * bytes_in_dtype["float64"] / (1024**3)
            )

        logger.debug(
            "Memory required for a single frequency channel: "
            + str(memory_singleton_chunk)
            + " GiB"
        )

        chunking_dims_sizes = {
            "frequency": img_xds.sizes["frequency"]
        }  # Need to know how many frequency channels there are.
        from astroviper._utils.data_partitioning import (
            calculate_data_chunking,
            get_thread_info,
        )

        if thread_info is None:
            thread_info = get_thread_info()
            logger.debug("Thread info " + str(thread_info))
        n_chunks = calculate_data_chunking(
            memory_singleton_chunk,
            chunking_dims_sizes,
            thread_info,
            constant_memory=0,
            tasks_per_thread=4,
        )["frequency"]
        logger.info(
            "Number of frequency chunks: "
            + str(n_chunks)
            + " frequency channels: "
            + str(chunking_dims_sizes)
        )

    # Make Parallel Coords
    parallel_coords = {}
    parallel_coords["frequency"] = make_parallel_coord(
        coord=frequency_coord, n_chunks=n_chunks
    )
    logger.info(
        "Number of frequency chunks: "
        + str(len(parallel_coords["frequency"]["data_chunks"]))
    )

    from xradio.image._util._zarr.zarr_low_level import (
        create_data_variable_meta_data,
    )

    if double_precission:
        from xradio.image._util._zarr.zarr_low_level import (
            image_data_variables_and_dims_double_precision as image_data_variables_and_dims,
        )
    else:
        from xradio.image._util._zarr.zarr_low_level import (
            image_data_variables_and_dims_single_precision as image_data_variables_and_dims,
        )

    xds_dims = dict(img_xds.dims)
    data_variables_and_dims_sel = {
        key: image_data_variables_and_dims[key] for key in data_variables
    }
    zarr_meta = create_data_variable_meta_data(
        image_name, data_variables_and_dims_sel, xds_dims, parallel_coords, compressor
    )

    sel_parms = {}
    sel_parms["intents"] = intents
    sel_parms["fields"] = None

    input_parms = {}
    input_parms["grid_params"] = grid_params
    input_parms["zarr_meta"] = zarr_meta
    input_parms["to_disk"] = True
    input_parms["polarization"] = polarization_coord.data
    input_parms["time"] = [0]
    input_parms["compressor"] = compressor
    input_parms["image_file"] = image_name
    input_parms["input_data_store"] = ps_name
    input_parms["data_group"] = data_group

    from graphviper.graph_tools.coordinate_utils import (
        interpolate_data_coords_onto_parallel_coords,
    )

    node_task_data_mapping = interpolate_data_coords_onto_parallel_coords(
        parallel_coords, ps
    )

    # Create Map Graph
    viper_graph = map(
        input_data=ps,
        node_task_data_mapping=node_task_data_mapping,
        node_task=_make_image,
        input_params=input_parms,
        in_memory_compute=False,
    )

    input_params = {}

    if workflow_ochestrator == "dask":
        viper_graph = reduce(
            viper_graph, _combine_return_data_frames, input_params, mode="tree"
        )
        # Compute cube
        dask_graph = generate_dask_workflow(viper_graph)
        # dask.visualize(dask_graph,filename='cube_imaging.png')
        return_dict = dask.compute(dask_graph)[0]
        zarr.consolidate_metadata(image_name)
        return return_dict
    elif workflow_ochestrator == "airflow":
        viper_graph = reduce(
            viper_graph, _combine_return_data_frames, input_params, mode="single_node"
        )
        generate_airflow_workflow(
            viper_graph,
            dag_id="0",
            schedule_interval=None,
            filename=os.path.join(airflow_dag_folder, "cube_imaging_airflow.py"),
            dag_name="cube_imaging",
        )

    # return parallel_coords


def _combine_return_data_frames(input_data, input_parms):
    import pandas as pd

    combined_df = pd.DataFrame()

    for df in input_data:
        combined_df = pd.concat([combined_df, df], ignore_index=True)

    return combined_df



import numpy as np
import scipy
from scipy import constants
from numba import jit
import numba
import xarray as xr
from astroviper._domain._imaging._check_imaging_parameters import _check_grid_parms
from astroviper._domain._imaging._imaging_utils._mosaic_grid import _aperture_grid_jit

# from graphviper.parameter_checking.check_parms import check_sel_parms
from astroviper.utils.check_parms import check_parms, check_sel_parms
import copy


def _make_aperture_grid(
    ms_xds, gcf_xds, img_xds, vis_sel_parms, img_sel_parms, grid_parms
):
    # Deep copy so that inputs are not modified
    _vis_sel_parms = copy.deepcopy(vis_sel_parms)
    _img_sel_parms = copy.deepcopy(img_sel_parms)
    _grid_parms = copy.deepcopy(grid_parms)

    _img_sel_parms["overwrite"] = True
    assert _check_grid_parms(_grid_parms), "######### ERROR: grid_parms checking failed"

    ms_data_group_in, ms_data_group_out = check_sel_parms(
        ms_xds, _vis_sel_parms, skip_data_group_out=True
    )
    img_data_group_in, img_data_group_out = check_sel_parms(
        img_xds,
        _img_sel_parms,
        default_data_group_out={
            "mosaic": {
                "aperture": "APERTURE",
                "aperture_normalization": "APERTURE_NORMALIZATION",
            }
        },
    )

    # print(ms_data_group_in, ms_data_group_out)

    weight_imaging = ms_xds[ms_data_group_in["weight_imaging"]].values
    n_chan = weight_imaging.shape[2]

    if _grid_parms["chan_mode"] == "cube":
        n_imag_chan = n_chan
        chan_map = (np.arange(0, n_chan)).astype(int)
    else:  # continuum
        n_imag_chan = 1  # Making only one continuum image.
        chan_map = (np.zeros(n_chan)).astype(int)

    n_imag_pol = weight_imaging.shape[3]
    pol_map = (np.arange(0, n_imag_pol)).astype(int)

    n_uv = _grid_parms["image_size_padded"]
    delta_lm = _grid_parms["cell_size"]
    oversampling = gcf_xds.attrs["oversampling"]
    # _grid_parms['oversampling']

    _grid_parms["complex_grid"] = True
    if img_data_group_out["aperture"] not in img_xds:
        img_xds[img_data_group_out["aperture"]] = xr.DataArray(
            np.zeros((n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.complex128),
            dims=["frequency", "polarization", "u", "v"],
        )
        img_xds[img_data_group_out["aperture_normalization"]] = xr.DataArray(
            np.zeros((n_imag_chan, n_imag_pol), dtype=np.double),
            dims=["frequency", "polarization"],
        )

    grid = img_xds[img_data_group_out["aperture"]].values
    sum_weight = img_xds[img_data_group_out["aperture_normalization"]].values

    uvw = ms_xds[ms_data_group_in["uvw"]].values
    freq_chan = ms_xds.frequency.values
    imaging_weight = ms_xds[ms_data_group_in["weight_imaging"]].values

    # print(img_xds)
    cf_baseline_map = gcf_xds["CF_BASELINE_MAP"].values
    cf_chan_map = gcf_xds["CF_CHAN_MAP"].values
    cf_pol_map = gcf_xds["CF_POL_MAP"].values
    conv_kernel_convolved = gcf_xds["CONV_KERNEL_CONVOLVED"].values
    weight_support = gcf_xds["SUPPORT"].values
    phase_gradient = gcf_xds["PHASE_GRADIENT"].values

    #    import matplotlib.pyplot as plt
    #    plt.figure()
    #    plt.plot(img_xds[img_data_group_out["aperture_normalization"]].values)
    #    plt.title('1')

    _aperture_grid_jit(
        grid,
        sum_weight,
        uvw,
        freq_chan,
        chan_map,
        pol_map,
        cf_baseline_map,
        cf_chan_map,
        cf_pol_map,
        imaging_weight,
        conv_kernel_convolved,
        n_uv,
        delta_lm,
        weight_support,
        oversampling,
        phase_gradient,
    )


#    plt.figure()
#    plt.plot(img_xds[img_data_group_out["sum_weight"]].values)
#    plt.title('2')
#    plt.figure()
#    plt.imshow(np.real(phase_gradient[0,:,:]))
#    plt.colorbar()

#    plt.figure()
#    plt.imshow(np.imag(phase_gradient[0,:,:]))
#    plt.colorbar()
#    plt.show()
#
#    print(phase_gradient[0,:,:])
#
#    plt.figure()
#    plt.imshow(np.abs(conv_kernel_convolved[0,0,0,:,:]))
#    plt.colorbar()
#
#
#    a = np.real(np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(img_xds[img_data_group_out["aperture"]].values, axes=(2, 3)), axes=(2, 3)), axes=(2, 3)))
#
#    plt.figure()
#    plt.imshow(np.abs(a[80,0,:,:]))
#    plt.colorbar()
#
#    plt.figure()
#    plt.imshow(np.abs(grid[80,0,:,:]))
#    plt.colorbar()
#
#    s = np.sum(np.abs(imaging_weight),axis=(0,1))
#    plt.figure()
#    plt.plot(s)
#
#    plt.show()



# Simple 1D Cases
# Airy Disk dish, blockage, freq
# Gaussian halfWidth
# Poly
# Cos Polyp
# Inverse Poly coeff


# Formula for obscured airy pattern found in https://en.wikipedia.org/wiki/Airy_disk (see Obscured Airy pattern section)
# If ipower is 1 the voltage pattern is returned if ipower is 2 the primary beam is returned.
def _airy_disk(freq_chan, pol, pb_parms, grid_parms):
    """
    Does not yet handle beam squint
    dish_diameters : list of int
    blockage_diameters : list of int
    frequencies : list of number
    """

    import numpy as np
    import scipy.constants
    from scipy.special import jn

    cell = grid_parms["cell_size"]
    image_size = grid_parms["image_size"]
    image_center = grid_parms["image_center"]

    list_dish_diameters = pb_parms["list_dish_diameters"]
    list_blockage_diameters = pb_parms["list_blockage_diameters"]
    ipower = pb_parms["ipower"]

    c = scipy.constants.c  # 299792458
    k = (2 * np.pi * freq_chan) / c

    x = np.arange(-image_center[0], image_size[0] - image_center[0]) * cell[0]
    y = np.arange(-image_center[1], image_size[1] - image_center[1]) * cell[1]

    airy_disk_size = (
        image_size[0],
        image_size[1],
        len(freq_chan),
        1,
        len(list_blockage_diameters),
    )  # len(pol) is set initially to 1. For now, the PB is assumed the same. This will change.
    airy_disk = np.zeros(airy_disk_size)

    for i, (dish_diameter, blockage_diameter) in enumerate(
        zip(list_dish_diameters, list_blockage_diameters)
    ):
        aperture = dish_diameter / 2
        x_grid, y_grid = np.meshgrid(x, y, indexing="ij")

        # r_grid = (np.sin(np.sqrt(x_grid**2 + y_grid**2))[:,:,None]*k*aperture) #d0 x d1 x chan
        r_grid = (
            np.sqrt(x_grid**2 + y_grid**2)[:, :, None] * k * aperture
        )  # d0 x d1 x chan
        r_grid[image_center[0], image_center[1], :] = (
            1.0  # Avoid the 0/0 for the centre value.
        )

        if blockage_diameter == 0.0:
            airy_disk[:, :, :, 0, i] = (2.0 * jn(1, r_grid) / r_grid) ** ipower
        else:
            e = blockage_diameter / dish_diameter
            airy_disk[:, :, :, 0, i] = (
                (2.0 * jn(1, r_grid) / r_grid - 2.0 * e * jn(1, r_grid * e) / r_grid)
                / (1.0 - e**2)
            ) ** ipower

    airy_disk[image_center[0], image_center[1], :, 0, :] = 1.0  # Fix centre value
    airy_disk = np.tile(airy_disk, (1, 1, 1, len(pol), 1))

    return airy_disk


# Formula for obscured airy pattern found in casa6/casa5/code/synthesis/TransformMachines/PBMath1DAiry.cc/h
# If ipower is 1 the voltage pattern is returned if ipower is 2 the primary beam is returned.
def _casa_airy_disk(freq_chan, pol, pb_parms, grid_parms):
    """
    Does not yet handle beam squint
    dish_diameters : list of int
    blockage_diameters : list of int
    frequencies : list of number
    """

    import numpy as np
    import scipy.constants
    from scipy.special import jn

    cell = grid_parms["cell_size"]
    image_size = grid_parms["image_size"]
    image_center = grid_parms["image_center"]

    list_dish_diameters = pb_parms["list_dish_diameters"]
    list_blockage_diameters = pb_parms["list_blockage_diameters"]
    ipower = pb_parms["ipower"]

    c = scipy.constants.c  # 299792458
    k = (2 * np.pi * freq_chan) / c

    x = np.arange(-image_center[0], image_size[0] - image_center[0]) * cell[0]
    y = np.arange(-image_center[1], image_size[1] - image_center[1]) * cell[1]

    airy_disk_size = (
        image_size[0],
        image_size[1],
        len(freq_chan),
        1,
        len(list_blockage_diameters),
    )  # len(pol) is set initially to 1. For now, the PB is assumed the same. This will change.
    airy_disk = np.zeros(airy_disk_size)

    for i, (dish_diameter, blockage_diameter) in enumerate(
        zip(list_dish_diameters, list_blockage_diameters)
    ):
        aperture = dish_diameter / 2
        x_grid, y_grid = np.meshgrid(x, y, indexing="ij")

        # r_grid = (np.sin(np.sqrt(x_grid**2 + y_grid**2))[:,:,None]*k*aperture) #d0 x d1 x chan
        r_grid = (
            np.sqrt(x_grid**2 + y_grid**2)[:, :, None] * k * aperture
        )  # d0 x d1 x chan
        r_grid[image_center[0], image_center[1], :] = (
            1.0  # Avoid the 0/0 for the centre value.
        )

        if blockage_diameter == 0.0:
            airy_disk[:, :, :, 0, i] = (2.0 * jn(1, r_grid) / r_grid) ** ipower
        else:
            area_ratio = (dish_diameter / blockage_diameter) ** 2
            length_ratio = dish_diameter / blockage_diameter
            airy_disk[:, :, :, 0, i] = (
                (
                    area_ratio * 2.0 * jn(1, r_grid) / r_grid
                    - 2.0 * jn(1, r_grid * length_ratio) / (r_grid * length_ratio)
                )
                / (area_ratio - 1.0)
            ) ** ipower

    airy_disk[image_center[0], image_center[1], :, 0, :] = 1.0  # Fix centre value
    airy_disk = np.tile(airy_disk, (1, 1, 1, len(pol), 1))

    return airy_disk


# Functions used during the creatiuon of the gridding convolution functions.
# Formula for obscured airy pattern found in https://en.wikipedia.org/wiki/Airy_disk (see Obscured Airy pattern section)
# If ipower is 1 the voltage pattern is returned if ipower is 2 the primary beam is returned.
def _airy_disk_rorder(freq_chan, pol, pb_parms, grid_parms):
    """
    Does not yet handle beam squint
    dish_diameters : list of int
    blockage_diameters : list of int
    frequencies : list of number
    """

    import numpy as np
    import scipy.constants
    from scipy.special import jn

    cell = grid_parms["cell_size"]
    image_size = grid_parms["image_size"]
    image_center = grid_parms["image_center"]

    list_dish_diameters = pb_parms["list_dish_diameters"]
    list_blockage_diameters = pb_parms["list_blockage_diameters"]
    ipower = pb_parms["ipower"]

    c = scipy.constants.c  # 299792458
    k = (2 * np.pi * freq_chan) / c

    x = np.arange(-image_center[0], image_size[0] - image_center[0]) * cell[0]
    y = np.arange(-image_center[1], image_size[1] - image_center[1]) * cell[1]

    airy_disk_size = (
        len(list_blockage_diameters),
        len(freq_chan),
        1,
        image_size[0],
        image_size[1],
    )  # len(pol) is set initially to 1. For now, the PB is assumed the same. This will change.
    airy_disk = np.zeros(airy_disk_size)

    for i, (dish_diameter, blockage_diameter) in enumerate(
        zip(list_dish_diameters, list_blockage_diameters)
    ):
        aperture = dish_diameter / 2
        x_grid, y_grid = np.meshgrid(x, y, indexing="ij")

        # r_grid = (np.sin(np.sqrt(x_grid**2 + y_grid**2))[:,:,None]*k*aperture) #d0 x d1 x chan
        r_grid = np.moveaxis(
            (np.sqrt(x_grid**2 + y_grid**2)[:, :, None] * k * aperture), 2, 0
        )  # chan x d0 x d1
        r_grid[:, image_center[0], image_center[1]] = (
            1.0  # Avoid the 0/0 for the centre value.
        )

        if blockage_diameter == 0.0:
            airy_disk[i, :, 0, :, :] = (2.0 * jn(1, r_grid) / r_grid) ** ipower
        else:
            e = blockage_diameter / dish_diameter
            airy_disk[i, :, 0, :, :] = (
                (2.0 * jn(1, r_grid) / r_grid - 2.0 * e * jn(1, r_grid * e) / r_grid)
                / (1.0 - e**2)
            ) ** ipower

    airy_disk[:, :, 0, image_center[0], image_center[1]] = 1.0  # Fix centre value
    # airy_disk[airy_disk<pb_limit] = 0.0
    airy_disk = np.tile(airy_disk, (1, 1, len(pol), 1, 1))

    return airy_disk


# Formula for obscured airy pattern found in casa6/casa5/code/synthesis/TransformMachines/PBMath1DAiry.cc/h
# If ipower is 1 the voltage pattern is returned if ipower is 2 the primary beam is returned.
def _casa_airy_disk_rorder(freq_chan, pol, pb_parms, grid_parms):
    """
    Does not yet handle beam squint
    dish_diameters : list of int
    blockage_diameters : list of int
    frequencies : list of number
    """

    import numpy as np
    import scipy.constants
    from scipy.special import jn

    cell = grid_parms["cell_size"]
    image_size = grid_parms["image_size"]
    image_center = grid_parms["image_center"]

    list_dish_diameters = pb_parms["list_dish_diameters"]
    list_blockage_diameters = pb_parms["list_blockage_diameters"]
    ipower = pb_parms["ipower"]

    c = scipy.constants.c  # 299792458
    k = (2 * np.pi * freq_chan) / c

    x = np.arange(-image_center[0], image_size[0] - image_center[0]) * cell[0]
    y = np.arange(-image_center[1], image_size[1] - image_center[1]) * cell[1]

    airy_disk_size = (
        len(list_blockage_diameters),
        len(freq_chan),
        1,
        image_size[0],
        image_size[1],
    )  # len(pol) is set initially to 1. For now, the PB is assumed the same. This will change.
    airy_disk = np.zeros(airy_disk_size)

    for i, (dish_diameter, blockage_diameter) in enumerate(
        zip(list_dish_diameters, list_blockage_diameters)
    ):
        aperture = dish_diameter / 2
        x_grid, y_grid = np.meshgrid(x, y, indexing="ij")

        # r_grid = (np.sin(np.sqrt(x_grid**2 + y_grid**2))[:,:,None]*k*aperture) #d0 x d1 x chan
        r_grid = np.moveaxis(
            (np.sqrt(x_grid**2 + y_grid**2)[:, :, None] * k * aperture), 2, 0
        )  # chan x d0 x d1
        r_grid[:, image_center[0], image_center[1]] = (
            1.0  # Avoid the 0/0 for the centre value.
        )

        if blockage_diameter == 0.0:
            airy_disk[i, :, 0, :, :] = (2.0 * jn(1, r_grid) / r_grid) ** ipower
        else:
            area_ratio = (dish_diameter / blockage_diameter) ** 2
            length_ratio = dish_diameter / blockage_diameter
            airy_disk[i, :, 0, :, :] = (
                (
                    area_ratio * 2.0 * jn(1, r_grid) / r_grid
                    - 2.0 * jn(1, r_grid * length_ratio) / (r_grid * length_ratio)
                )
                / (area_ratio - 1.0)
            ) ** ipower

    airy_disk[:, :, 0, image_center[0], image_center[1]] = 1.0  # Fix centre value
    airy_disk = np.tile(airy_disk, (1, 1, len(pol), 1, 1))

    return airy_disk



import toolviper.utils.logger as logger

bytes_in_dtype = {
    "float32": 4,
    "float64": 8,
    "double": 8,
    "complex64": 8,
    "complex128": 16,
}


def get_thread_info(client=None):

    if client is None:
        try:
            from distributed import Client

            client = Client.current()
        except:  # Using default Dask schedular.
            import psutil

            cpu_cores = psutil.cpu_count()
            total_memory = psutil.virtual_memory().total / (1024**3)
            # print(cpu_cores,total_memory)
            thread_info = {
                "n_threads": cpu_cores,
                "memory_per_thread": total_memory / cpu_cores,
            }
            return thread_info

    memory_per_thread = -1
    n_threads = 0

    # client.cluster only exists for LocalCluster
    if client.cluster == None:
        worker_items = client.scheduler_info()["workers"].items()
    else:
        worker_items = client.cluster.scheduler_info["workers"].items()

    for worker_name, worker in worker_items:
        temp_memory_per_thread = (worker["memory_limit"] / worker["nthreads"]) / (
            1024**3
        )
        n_threads = n_threads + worker["nthreads"]
        if (memory_per_thread == -1) or (memory_per_thread > temp_memory_per_thread):
            memory_per_thread = temp_memory_per_thread

    thread_info = {"n_threads": n_threads, "memory_per_thread": memory_per_thread}
    return thread_info


def prime_factors(n):
    i = 2
    factors = []
    while i * i <= n:
        if n % i:
            i += 1
        else:
            n //= i
            factors.append(i)
    if n > 1:
        factors.append(n)
    return factors


def calculate_data_chunking(
    memory_singleton_chunk,
    chunking_dims_sizes,
    thread_info,
    constant_memory=0,
    tasks_per_thread=4,
):

    import numpy as np

    n_total_dims = np.prod(np.array(list(chunking_dims_sizes.values())))

    n_mem_chunks = int(
        np.ceil(
            n_total_dims
            / (
                (thread_info["memory_per_thread"] - constant_memory)
                / memory_singleton_chunk
            )
        )
    )
    assert n_mem_chunks <= n_total_dims, (
        "Not enough cluster memory per thread. Need at least "
        + str(memory_singleton_chunk)
        + " GiB but only "
        + str(thread_info["memory_per_thread"])
        + " GiB available."
    )

    n_graph_chunks = int(thread_info["n_threads"] * tasks_per_thread)
    if n_graph_chunks > n_total_dims:
        n_graph_chunks = n_total_dims

    if n_mem_chunks > n_graph_chunks:
        n_chunks = n_mem_chunks
    else:
        n_chunks = n_graph_chunks

    logger.debug(
        "Suggest n_chunks: "
        + str(n_chunks)
        + ", n_mem_chunks: "
        + str(n_mem_chunks)
        + ", n_graph_chunks: "
        + str(n_graph_chunks)
    )

    dims_sizes = []
    dims_names = []

    for dim_name in sorted(chunking_dims_sizes, key=chunking_dims_sizes.get):
        dims_sizes.append(chunking_dims_sizes[dim_name])
        dims_names.append(dim_name)

    dims_sizes_arr = np.array(dims_sizes)
    dims_names_arr = np.array(dims_names)

    if len(dims_names_arr) == 1:
        n_chunks_dict = dict(zip([dims_names_arr[0]], [n_chunks]))
        return n_chunks_dict
    else:
        factors = prime_factors(n_chunks)

        found_factors = False

        while not found_factors:

            if len(factors) > len(chunking_dims_sizes):

                n_reduce = len(factors) - len(chunking_dims_sizes)

                for i in range(n_reduce):
                    if factors[0] * factors[-(i + 1)] < dims_sizes_arr[-(i + 1)]:
                        new_factors = factors[1:]
                        new_factors[-(i + 1)] = factors[0] * factors[-(i + 1)]
                        factors = new_factors

            elif len(factors) < len(chunking_dims_sizes):
                while len(factors) < len(chunking_dims_sizes):
                    n_chunks = n_chunks + 1
                    factors = prime_factors(n_chunks_inc)
            else:
                found_factors = True

        # print(len(chunking_dims_sizes), len(factors), factors)

        n_chunks_dict = dict(zip(dims_names_arr, factors))
        return n_chunks_dict



import numpy as np
import scipy
from scipy import constants
from numba import jit
import numba
import xarray as xr
from astroviper._domain._imaging._check_imaging_parameters import _check_grid_parms
from astroviper._domain._imaging._imaging_utils._mosaic_grid import _mosaic_grid_jit
from astroviper._domain._imaging._imaging_utils._standard_grid import _standard_grid_jit

# from graphviper.parameter_checking.check_parms import check_sel_parms
from astroviper.utils.check_parms import check_parms, check_sel_parms
import copy


def _make_visibility_grid(
    ms_xds, gcf_xds, img_xds, vis_sel_parms, img_sel_parms, grid_parms
):
    # Deep copy so that inputs are not modified
    _vis_sel_parms = copy.deepcopy(vis_sel_parms)
    _img_sel_parms = copy.deepcopy(img_sel_parms)
    _grid_parms = copy.deepcopy(grid_parms)

    _img_sel_parms["overwrite"] = True
    assert _check_grid_parms(_grid_parms), "######### ERROR: grid_parms checking failed"

    ms_data_group_in, ms_data_group_out = check_sel_parms(
        ms_xds, _vis_sel_parms, skip_data_group_out=True
    )
    img_data_group_in, img_data_group_out = check_sel_parms(
        img_xds,
        _img_sel_parms,
        default_data_group_out={
            "mosaic": {
                "visibility": "VISIBILITY",
                "visibility_normalization": "VISIBILITY_NORMALIZATION",
            }
        },
    )

    # print(ms_data_group_in, ms_data_group_out)

    weight_imaging = ms_xds[ms_data_group_in["weight_imaging"]].values
    n_chan = weight_imaging.shape[2]

    if _grid_parms["chan_mode"] == "cube":
        n_imag_chan = n_chan
        chan_map = (np.arange(0, n_chan)).astype(int)
    else:  # continuum
        n_imag_chan = 1  # Making only one continuum image.
        chan_map = (np.zeros(n_chan)).astype(int)

    n_imag_pol = weight_imaging.shape[3]
    pol_map = (np.arange(0, n_imag_pol)).astype(int)

    n_uv = _grid_parms["image_size_padded"]
    delta_lm = _grid_parms["cell_size"]
    oversampling = gcf_xds.attrs["oversampling"]

    _grid_parms["complex_grid"] = True
    if img_data_group_out["visibility"] not in img_xds:
        if _grid_parms["complex_grid"]:
            img_xds[img_data_group_out["visibility"]] = xr.DataArray(
                np.zeros(
                    (n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.complex128
                ),
                dims=["frequency", "polarization", "u", "v"],
            )

        else:
            img_xds[img_data_group_out["visibility"]] = xr.DataArray(
                np.zeros((n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.double),
                dims=["frequency", "polarization", "u", "v"],
            )
        img_xds[img_data_group_out["visibility_normalization"]] = xr.DataArray(
            np.zeros((n_imag_chan, n_imag_pol), dtype=np.double),
            dims=["frequency", "polarization"],
        )

    grid = img_xds[img_data_group_out["visibility"]].values
    sum_weight = img_xds[img_data_group_out["visibility_normalization"]].values

    vis_data = ms_xds[ms_data_group_in["correlated_data"]].values
    uvw = ms_xds[ms_data_group_in["uvw"]].values
    freq_chan = ms_xds.frequency.values
    imaging_weight = ms_xds[ms_data_group_in["weight_imaging"]].values

    do_psf = False

    # print(img_xds)
    cf_baseline_map = gcf_xds["CF_BASELINE_MAP"].values
    cf_chan_map = gcf_xds["CF_CHAN_MAP"].values
    cf_pol_map = gcf_xds["CF_POL_MAP"].values
    conv_kernel = gcf_xds["CONV_KERNEL"].values
    weight_support = gcf_xds["SUPPORT"].values
    phase_gradient = gcf_xds["PHASE_GRADIENT"].values

    _mosaic_grid_jit(
        grid,
        sum_weight,
        do_psf,
        vis_data,
        uvw,
        freq_chan,
        chan_map,
        pol_map,
        cf_baseline_map,
        cf_chan_map,
        cf_pol_map,
        imaging_weight,
        conv_kernel,
        n_uv,
        delta_lm,
        weight_support,
        oversampling,
        phase_gradient,
    )


def _make_visibility_grid_single_field(
    ms_xds, cgk_1D, img_xds, vis_sel_parms, img_sel_parms, grid_parms
):
    # Deep copy so that inputs are not modified
    _vis_sel_parms = copy.deepcopy(vis_sel_parms)
    _img_sel_parms = copy.deepcopy(img_sel_parms)
    _grid_parms = copy.deepcopy(grid_parms)

    _img_sel_parms["overwrite"] = True
    assert _check_grid_parms(_grid_parms), "######### ERROR: grid_parms checking failed"

    ms_data_group_in, ms_data_group_out = check_sel_parms(
        ms_xds, _vis_sel_parms, skip_data_group_out=True
    )
    img_data_group_in, img_data_group_out = check_sel_parms(
        img_xds,
        _img_sel_parms,
        default_data_group_out={
            "mosaic": {
                "visibility": "VISIBILITY",
                "visibility_normalization": "VISIBILITY_NORMALIZATION",
            }
        },
    )

    # print(ms_data_group_in, ms_data_group_out)

    weight_imaging = ms_xds[ms_data_group_in["weight_imaging"]].values
    n_chan = weight_imaging.shape[2]

    if _grid_parms["chan_mode"] == "cube":
        n_imag_chan = n_chan
        chan_map = (np.arange(0, n_chan)).astype(int)
    else:  # continuum
        n_imag_chan = 1  # Making only one continuum image.
        chan_map = (np.zeros(n_chan)).astype(int)

    n_imag_pol = weight_imaging.shape[3]
    pol_map = (np.arange(0, n_imag_pol)).astype(int)

    n_uv = _grid_parms["image_size_padded"]
    delta_lm = _grid_parms["cell_size"]

    _grid_parms["complex_grid"] = True
    if img_data_group_out["visibility"] not in img_xds:
        if _grid_parms["complex_grid"]:
            img_xds[img_data_group_out["visibility"]] = xr.DataArray(
                np.zeros(
                    (n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.complex128
                ),
                dims=["frequency", "polarization", "u", "v"],
            )

        else:
            img_xds[img_data_group_out["visibility"]] = xr.DataArray(
                np.zeros((n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.double),
                dims=["frequency", "polarization", "u", "v"],
            )
        img_xds[img_data_group_out["visibility_normalization"]] = xr.DataArray(
            np.zeros((n_imag_chan, n_imag_pol), dtype=np.double),
            dims=["frequency", "polarization"],
        )

    grid = img_xds[img_data_group_out["visibility"]].values
    sum_weight = img_xds[img_data_group_out["visibility_normalization"]].values

    vis_data = ms_xds[ms_data_group_in["correlated_data"]].values
    uvw = ms_xds[ms_data_group_in["uvw"]].values
    freq_chan = ms_xds.frequency.values
    imaging_weight = ms_xds[ms_data_group_in["weight_imaging"]].values

    do_psf = False
    do_imaging_weight = False

    _standard_grid_jit(
        grid,
        sum_weight,
        do_psf,
        do_imaging_weight,
        vis_data,
        uvw,
        freq_chan,
        chan_map,
        pol_map,
        imaging_weight,
        cgk_1D,
        n_uv,
        delta_lm,
        support=7,
        oversampling=100,
    )



from numba import jit
import numpy as np
import math

# from numba import gdb


def ndim_list(shape):
    return [ndim_list(shape[1:]) if len(shape) > 1 else None for _ in range(shape[0])]


@jit(nopython=True, cache=True, nogil=True)
def _aperture_grid_jit(
    grid,
    sum_weight,
    uvw,
    freq_chan,
    chan_map,
    pol_map,
    cf_baseline_map,
    cf_chan_map,
    cf_pol_map,
    imaging_weight,
    conv_kernel_convolved,
    n_uv,
    delta_lm,
    weight_support,
    oversampling,
    phase_gradient,
):
    c = 299792458.0
    uv_scale = np.zeros((2, len(freq_chan)), dtype=np.double)
    uv_scale[0, :] = -(freq_chan * delta_lm[0] * n_uv[0]) / c
    uv_scale[1, :] = -(freq_chan * delta_lm[1] * n_uv[1]) / c

    uv_center = n_uv // 2

    n_time = uvw.shape[0]
    n_baseline = uvw.shape[1]

    n_chan = len(chan_map)
    n_pol = len(pol_map)

    n_u = n_uv[0]
    n_v = n_uv[1]

    u_center = uv_center[0]
    v_center = uv_center[1]

    max_support_center = np.max(weight_support)

    conv_v_center = conv_kernel_convolved.shape[-1] // 2
    conv_u_center = conv_kernel_convolved.shape[-2] // 2

    # print(phase_gradient.shape)
    # print(conv_kernel_convolved.shape)

    conv_kernel_convolved_phase_gradient = (
        conv_kernel_convolved * phase_gradient[0, :, :]
    )

    for i_time in range(n_time):
        for i_baseline in range(n_baseline):
            cf_baseline = cf_baseline_map[i_baseline]
            for i_chan in range(n_chan):
                cf_chan = cf_chan_map[i_chan]
                a_chan = chan_map[i_chan]
                u = uvw[i_time, i_baseline, 0] * uv_scale[0, i_chan]
                v = uvw[i_time, i_baseline, 1] * uv_scale[1, i_chan]

                if ~np.isnan(u) and ~np.isnan(v):
                    u_pos = u + uv_center[0]
                    v_pos = v + uv_center[1]

                    # Do not use numpy round
                    u_center_indx = int(u_pos + 0.5)
                    v_center_indx = int(v_pos + 0.5)

                    if (
                        (u_center_indx + max_support_center < n_u)
                        and (v_center_indx + max_support_center < n_v)
                        and (u_center_indx - max_support_center >= 0)
                        and (v_center_indx - max_support_center >= 0)
                    ):
                        # u_offset = u_center_indx - u_pos
                        # u_center_offset_indx = math.floor(u_offset * oversampling[0] + 0.5)
                        # v_offset = v_center_indx - v_pos
                        # v_center_offset_indx = math.floor(v_offset * oversampling[1] + 0.5)

                        for i_pol in range(n_pol):
                            weighted_data = imaging_weight[
                                i_time, i_baseline, i_chan, i_pol
                            ]

                            if ~np.isnan(weighted_data) and (weighted_data != 0.0):
                                cf_pol = cf_pol_map[i_pol]
                                a_pol = pol_map[i_pol]
                                norm = 0.0

                                """
                                support = weight_support[cf_baseline,cf_chan,cf_pol,:]
                                #support = np.array([13,13])
                                support_center = support // 2
                                start_support = - support_center
                                end_support = support - support_center # end_support is larger by 1 so that python range() gives correct indices
                                """

                                support_u = weight_support[
                                    cf_baseline, cf_chan, cf_pol, 0
                                ]
                                support_v = weight_support[
                                    cf_baseline, cf_chan, cf_pol, 1
                                ]

                                support_center_u = support_u // 2
                                support_center_v = support_v // 2

                                start_support_u = -support_center_u
                                start_support_v = -support_center_v

                                end_support_u = support_u - support_center_u
                                end_support_v = support_v - support_center_v

                                for i_v in range(start_support_v, end_support_v):
                                    v_indx = v_center + i_v
                                    cf_v_indx = oversampling[1] * i_v + conv_v_center

                                    for i_u in range(start_support_u, end_support_u):
                                        u_indx = u_center + i_u
                                        cf_u_indx = (
                                            oversampling[0] * i_u + conv_u_center
                                        )

                                        conv = conv_kernel_convolved_phase_gradient[
                                            cf_baseline,
                                            cf_chan,
                                            cf_pol,
                                            cf_u_indx,
                                            cf_v_indx,
                                        ]

                                        grid[a_chan, a_pol, u_indx, v_indx] = (
                                            grid[a_chan, a_pol, u_indx, v_indx]
                                            + conv * weighted_data
                                        )
                                        norm = norm + conv

                                # print("The norm",norm)
                                sum_weight[a_chan, a_pol] = sum_weight[
                                    a_chan, a_pol
                                ] + weighted_data * np.real(norm)

    return


def _mosaic_grid_numpy_wrap(
    vis_data,
    uvw,
    imaging_weight,
    field,
    cf_baseline_map,
    cf_chan_map,
    cf_pol_map,
    conv_kernel,
    weight_support,
    phase_gradient,
    freq_chan,
    grid_parms,
):
    # print('imaging_weight ', imaging_weight.shape)
    import time

    n_chan = imaging_weight.shape[2]
    if grid_parms["chan_mode"] == "cube":
        n_imag_chan = n_chan
        chan_map = (np.arange(0, n_chan)).astype(np.int)
    else:  # continuum
        n_imag_chan = 1  # Making only one continuum image.
        chan_map = (np.zeros(n_chan)).astype(np.int)

    n_imag_pol = imaging_weight.shape[3]
    pol_map = (np.arange(0, n_imag_pol)).astype(np.int)

    n_uv = grid_parms["image_size_padded"]
    delta_lm = grid_parms["cell_size"]
    oversampling = grid_parms["oversampling"]

    if grid_parms["complex_grid"]:
        grid = np.zeros(
            (n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.complex128
        )
    else:
        grid = np.zeros((n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.double)
    sum_weight = np.zeros((n_imag_chan, n_imag_pol), dtype=np.double)

    do_psf = grid_parms["do_psf"]
    field_id = grid_parms["field_id"]

    # print('vis_data', vis_data.shape , 'grid ', grid.shape, 'sum_weight', sum_weight.shape, 'cf_chan_map ', cf_chan_map.shape, ' cf_baseline_map', cf_baseline_map.shape, 'cf_pol_map', cf_pol_map.shape, ' conv_kernel',  conv_kernel.shape, 'phase_gradient', phase_gradient.shape, 'field', field.shape,  )

    # start = time.time()
    _aperture_grid_jit(
        grid,
        sum_weight,
        do_psf,
        vis_data,
        uvw,
        freq_chan,
        chan_map,
        pol_map,
        cf_baseline_map,
        cf_chan_map,
        cf_pol_map,
        imaging_weight,
        conv_kernel,
        n_uv,
        delta_lm,
        weight_support,
        oversampling,
        field,
        field_id,
        phase_gradient,
    )
    # time_to_grid = time.time() - start
    # print("time to grid ", time_to_grid)

    return grid, sum_weight


def _mosaic_psf_grid_numpy_wrap(
    uvw,
    imaging_weight,
    field,
    cf_baseline_map,
    cf_chan_map,
    cf_pol_map,
    conv_kernel,
    weight_support,
    phase_gradient,
    freq_chan,
    grid_parms,
):
    # print('imaging_weight ', imaging_weight.shape)
    import time

    n_chan = imaging_weight.shape[2]
    if grid_parms["chan_mode"] == "cube":
        n_imag_chan = n_chan
        chan_map = (np.arange(0, n_chan)).astype(np.int)
    else:  # continuum
        n_imag_chan = 1  # Making only one continuum image.
        chan_map = (np.zeros(n_chan)).astype(np.int)

    n_imag_pol = imaging_weight.shape[3]
    pol_map = (np.arange(0, n_imag_pol)).astype(np.int)

    n_uv = grid_parms["image_size_padded"]
    delta_lm = grid_parms["cell_size"]
    oversampling = grid_parms["oversampling"]

    if grid_parms["complex_grid"]:
        grid = np.zeros(
            (n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.complex128
        )
    else:
        grid = np.zeros((n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.double)
    sum_weight = np.zeros((n_imag_chan, n_imag_pol), dtype=np.double)

    do_psf = grid_parms["do_psf"]
    field_id = grid_parms["field_id"]

    # print('vis_data', vis_data.shape , 'grid ', grid.shape, 'sum_weight', sum_weight.shape, 'cf_chan_map ', cf_chan_map.shape, ' cf_baseline_map', cf_baseline_map.shape, 'cf_pol_map', cf_pol_map.shape, ' conv_kernel',  conv_kernel.shape, 'phase_gradient', phase_gradient.shape, 'field', field.shape,  )

    vis_data = np.zeros((1, 1, 1, 1), dtype=np.bool)
    # start = time.time()
    _mosaic_grid_jit(
        grid,
        sum_weight,
        do_psf,
        vis_data,
        uvw,
        freq_chan,
        chan_map,
        pol_map,
        cf_baseline_map,
        cf_chan_map,
        cf_pol_map,
        imaging_weight,
        conv_kernel,
        n_uv,
        delta_lm,
        weight_support,
        oversampling,
        field,
        field_id,
        phase_gradient,
    )
    # time_to_grid = time.time() - start
    # print("time to grid ", time_to_grid)

    return grid, sum_weight


# Important changes to be made https://github.com/numba/numba/issues/4261
# debug=True and gdb()
@jit(nopython=True, cache=True, nogil=True)
def _mosaic_grid_jit(
    grid,
    sum_weight,
    do_psf,
    vis_data,
    uvw,
    freq_chan,
    chan_map,
    pol_map,
    cf_baseline_map,
    cf_chan_map,
    cf_pol_map,
    imaging_weight,
    conv_kernel,
    n_uv,
    delta_lm,
    weight_support,
    oversampling,
    phase_gradient,
):
    c = 299792458.0
    uv_scale = np.zeros((2, len(freq_chan)), dtype=np.double)
    uv_scale[0, :] = -(freq_chan * delta_lm[0] * n_uv[0]) / c
    uv_scale[1, :] = -(freq_chan * delta_lm[1] * n_uv[1]) / c

    uv_center = n_uv // 2

    n_time = uvw.shape[0]
    n_baseline = uvw.shape[1]
    n_chan = len(chan_map)
    n_pol = len(pol_map)

    n_u = n_uv[0]
    n_v = n_uv[1]

    u_center = uv_center[0]
    v_center = uv_center[1]

    max_support_center = np.max(weight_support)

    conv_v_center = conv_kernel.shape[-1] // 2
    conv_u_center = conv_kernel.shape[-2] // 2

    conv_size = np.array(conv_kernel.shape[-2:])

    # print('conv_size',conv_size)

    # print('sizes ',conv_kernel.shape, conv_u_center, conv_v_center)

    # print(phase_gradient.shape)
    # print(conv_kernel_convolved.shape)

    conv_kernel_phase_gradient = conv_kernel * phase_gradient[0, :, :]

    for i_time in range(n_time):
        for i_baseline in range(n_baseline):
            cf_baseline = cf_baseline_map[i_baseline]
            for i_chan in range(n_chan):
                cf_chan = cf_chan_map[i_chan]
                a_chan = chan_map[i_chan]
                u = uvw[i_time, i_baseline, 0] * uv_scale[0, i_chan]
                v = uvw[i_time, i_baseline, 1] * uv_scale[1, i_chan]

                if ~np.isnan(u) and ~np.isnan(v):
                    u_pos = u + uv_center[0]
                    v_pos = v + uv_center[1]

                    # Do not use numpy round
                    u_center_indx = int(u_pos + 0.5)
                    v_center_indx = int(v_pos + 0.5)

                    if (
                        (u_center_indx + max_support_center < n_u)
                        and (v_center_indx + max_support_center < n_v)
                        and (u_center_indx - max_support_center >= 0)
                        and (v_center_indx - max_support_center >= 0)
                    ):
                        u_offset = u_center_indx - u_pos
                        u_center_offset_indx = (
                            math.floor(u_offset * oversampling[0] + 0.5) + conv_u_center
                        )
                        v_offset = v_center_indx - v_pos
                        v_center_offset_indx = (
                            math.floor(v_offset * oversampling[1] + 0.5) + conv_v_center
                        )

                        for i_pol in range(n_pol):
                            if do_psf:
                                weighted_data = imaging_weight[
                                    i_time, i_baseline, i_chan, i_pol
                                ]
                            else:
                                weighted_data = (
                                    vis_data[i_time, i_baseline, i_chan, i_pol]
                                    * imaging_weight[i_time, i_baseline, i_chan, i_pol]
                                )

                            if ~np.isnan(weighted_data) and (weighted_data != 0.0):
                                cf_pol = cf_pol_map[i_pol]
                                a_pol = pol_map[i_pol]
                                norm = 0.0

                                """
                                support = weight_support[cf_baseline,cf_chan,cf_pol,:]
                                #support = np.array([13,13])
                                support_center = support // 2
                                start_support = - support_center
                                end_support = support - support_center # end_support is larger by 1 so that python range() gives correct indices
                                """

                                support_u = weight_support[
                                    cf_baseline, cf_chan, cf_pol, 0
                                ]
                                support_v = weight_support[
                                    cf_baseline, cf_chan, cf_pol, 1
                                ]

                                support_center_u = support_u // 2
                                support_center_v = support_v // 2

                                start_support_u = -support_center_u
                                start_support_v = -support_center_v

                                end_support_u = support_u - support_center_u
                                end_support_v = support_v - support_center_v

                                # print(support)
                                ###############
                                #                                resized_conv_size = (support  + 1)*oversampling
                                #                                start_indx = conv_size//2 - resized_conv_size//2
                                #                                end_indx = start_indx + resized_conv_size
                                #                                normalize_factor = np.real(np.sum(conv_kernel[cf_baseline,cf_chan,cf_pol,start_indx[0]:end_indx[0],start_indx[1]:end_indx[1]])/(oversampling[0]*oversampling[1]))
                                #
                                #                                conv_kernel_phase_gradient = conv_kernel*phase_gradient[field[i_time],:,:]/normalize_factor
                                #                                print(normalize_factor)
                                ##############

                                for i_v in range(start_support_v, end_support_v):
                                    v_indx = v_center_indx + i_v
                                    cf_v_indx = (
                                        oversampling[1] * i_v + v_center_offset_indx
                                    )

                                    for i_u in range(start_support_u, end_support_u):
                                        u_indx = u_center_indx + i_u
                                        cf_u_indx = (
                                            oversampling[0] * i_u + u_center_offset_indx
                                        )

                                        conv = conv_kernel_phase_gradient[
                                            cf_baseline,
                                            cf_chan,
                                            cf_pol,
                                            cf_u_indx,
                                            cf_v_indx,
                                        ]

                                        grid[a_chan, a_pol, u_indx, v_indx] = (
                                            grid[a_chan, a_pol, u_indx, v_indx]
                                            + conv * weighted_data
                                        )
                                        norm = norm + conv
                                if do_psf:
                                    sum_weight[a_chan, a_pol] = sum_weight[
                                        a_chan, a_pol
                                    ] + imaging_weight[
                                        i_time, i_baseline, i_chan, i_pol
                                    ] * np.real(
                                        norm
                                    )
                                else:
                                    sum_weight[a_chan, a_pol] = sum_weight[
                                        a_chan, a_pol
                                    ] + imaging_weight[
                                        i_time, i_baseline, i_chan, i_pol
                                    ] * np.real(
                                        norm**2
                                    )  # *np.real(norm**2)#* np.real(norm) #np.abs(norm**2) #**2 term is needed since the pb is in the image twice (one naturally and another from the gcf)

    return


def _tree_sum_list(list_to_sum):
    import dask.array as da

    while len(list_to_sum) > 1:
        new_list_to_sum = []
        for i in range(0, len(list_to_sum), 2):
            if i < len(list_to_sum) - 1:
                lazy = da.add(list_to_sum[i], list_to_sum[i + 1])
            else:
                lazy = list_to_sum[i]
            new_list_to_sum.append(lazy)
        list_to_sum = new_list_to_sum
    return list_to_sum



# ducting - code is complex and might fail after some time if parameters is wrong. Sensable values are also checked. Gives printout of all wrong parameters. Dirty images alone has 14 parameters.

import numpy as np
from .check_parms import check_parms, _check_dataset


def check_logger_parms(logger_parms):
    import numbers

    parms_passed = True
    arc_sec_to_rad = np.pi / (3600 * 180)

    if not (check_parms(logger_parms, "log_to_term", [bool], default=True)):
        parms_passed = False
    if not (check_parms(logger_parms, "log_to_file", [bool], default=False)):
        parms_passed = False
    if not (check_parms(logger_parms, "log_file", [str], default="viper_")):
        parms_passed = False
    if not (
        check_parms(
            logger_parms,
            "log_level",
            [str],
            default="INFO",
            acceptable_data=["DEBUG", "INFO", "WARNING", "ERROR"],
        )
    ):
        parms_passed = False

    return parms_passed


def check_worker_logger_parms(logger_parms):
    import numbers

    parms_passed = True
    arc_sec_to_rad = np.pi / (3600 * 180)

    if not (check_parms(logger_parms, "log_to_term", [bool], default=False)):
        parms_passed = False
    if not (check_parms(logger_parms, "log_to_file", [bool], default=False)):
        parms_passed = False
    if not (check_parms(logger_parms, "log_file", [str], default="viper_")):
        parms_passed = False
    if not (
        check_parms(
            logger_parms,
            "log_level",
            [str],
            default="INFO",
            acceptable_data=["DEBUG", "INFO", "WARNING", "ERROR"],
        )
    ):
        parms_passed = False

    return parms_passed



import xarray as xr


def _moment_max(input_data, input_parms):
    # print(len(input_data),input_data[0].dims,input_data[1].dims)
    # print(input_data)
    # print('********')

    for i, img_xds in enumerate(input_data):
        input_data[i] = img_xds.max(dim="frequency", keepdims=True)

    img_xds = xr.concat(input_data, dim="frequency")
    return img_xds.max(dim="frequency", keepdims=True)



import numpy as np
import scipy
from scipy import constants
from numba import jit
import numba
import xarray as xr
import copy

# silence NumbaPerformanceWarning
# import warnings
# from numba.errors import NumbaPerformanceWarning
# warnings.filterwarnings("ignore", category=NumbaPerformanceWarning) #Suppress  NumbaPerformanceWarning: '@' is faster on contiguous arrays warning. This happens for phasor_loop and apply_rotation_matrix functions.

# Based on CASACORE measures/Measures/UVWMachine.cc and CASA code/synthesis/TransformMachines2/FTMachine.cc::girarUVW


def _phase_shift_vis_ds(ms_xds, shift_parms, sel_parms):
    """
    Rotate uvw coordinates and phase rotate visibilities. For a joint mosaics rotation_parms['common_tangent_reprojection'] must be true.
    The specified phasedirection and field phase directions are assumed to be in the same frame.
    East-west arrays, emphemeris objects or objects within the nearfield are not supported.

    Parameters
    ----------
    mxds : xarray.core.dataset.Dataset
        Input vis.zarr multi dataset.
    shift_parms : dictionary
    shift_parms['new_phase_direction'] : list of number, length = 2, units = radians
       The phase direction to rotate to (right ascension and declination).
    shift_parms['common_tangent_reprojection']  : bool, default = True
       If true common tangent reprojection is used (should be true if a joint mosaic image is being created).
    shift_parms['single_precision'] : bool, default = True
       If shift_parms['single_precision'] is true then the output visibilities are cast from 128 bit complex to 64 bit complex. Mathematical operations are always done in double precision.
    sel_parms : dict
    sel_parms['data_group_in'] : dict, default = vis_dataset.data_group[0][0]
        Only the id has to be specified
        The names of the input data and uvw data variables.
        For example sel_parms['data_group_in'] = {'id':'1', 'data':'DATA','uvw':'UVW'}.
    sel_parms['data_group_out'] : dict, default = {**_vis_dataset.data_group[0],**{'id':str(new_data_group_id),'uvw':'UVW_ROT','data':'DATA_ROT','field_id':shift_parms['new_phase_direction']}}
        The names of the new data and uvw data variables that have been direction rotated.
        For example sel_parms['data_group_out'] = {'id':2,'data':'DATA_ROT','uvw':'UVW_ROT'}.
        The new_data_group_id is the next available id.
    sel_parms['overwrite'] : bool, default = False
    Returns
    -------
    psf_dataset : xarray.core.dataset.Dataset
    """

    # from graphviper.parameter_checking.check_parms import check_sel_parms
    from astroviper.utils.check_parms import check_parms, check_sel_parms

    _sel_parms = copy.deepcopy(sel_parms)
    _shift_parms = copy.deepcopy(shift_parms)

    assert _check_shift_parms(
        _shift_parms
    ), "######### ERROR: shift_parms checking failed"

    data_group_in, data_group_out = check_sel_parms(
        ms_xds,
        _sel_parms,
        default_data_group_out={
            "phase_shift": {"correlated_data": "VISIBILITY_SHIFT", "uvw": "UVW_SHIFT"}
        },
    )

    field_phase_direction = ms_xds[data_group_in["correlated_data"]].attrs[
        "field_and_source_xds"
    ]["FIELD_PHASE_CENTER"]
    uvw_rotmat, phase_rotation = calc_rotation_matrices(
        field_phase_direction, _shift_parms
    )

    # print("uvw_rotmat, phase_rotation",uvw_rotmat, phase_rotation)

    ms_xds[data_group_out["uvw"]] = xr.DataArray(
        ms_xds[data_group_in["uvw"]].values @ uvw_rotmat,
        dims=ms_xds[data_group_in["uvw"]].dims,
    )

    if _shift_parms["common_tangent_reprojection"]:
        end_slice = 2
    else:
        end_slice = 3

    phase_direction = (
        ms_xds[data_group_out["uvw"]][:, :, 0:end_slice].values
        @ phase_rotation[0:end_slice]
    )

    phasor = np.exp(
        2.0
        * 1j
        * np.pi
        * phase_direction[:, :, None, None]
        * ms_xds.frequency.values[None, None, :, None]
        / constants.c
    )  # phasor_ngcasa = - phasor_casa. Sign flip is due to CASA

    if _shift_parms["single_precision"]:
        # print("single precision")
        ms_xds[data_group_out["correlated_data"]] = (
            (phasor * ms_xds[data_group_in["correlated_data"]]).astype(np.complex64)
        ).astype(np.complex128)
    else:
        ms_xds[data_group_out["correlated_data"]] = (
            phasor * ms_xds[data_group_in["correlated_data"]]
        )

    ms_xds[data_group_out["correlated_data"]].attrs["phase_direction"] = shift_parms[
        "new_phase_direction"
    ]

    return _sel_parms["data_group_out"]


def calc_rotation_matrices(field_phase_direction, shift_parms):
    from scipy.spatial.transform import Rotation as R

    ra_image = shift_parms["new_phase_direction"].values[0]
    dec_image = shift_parms["new_phase_direction"].values[1]

    rotmat_new_phase_direction = R.from_euler(
        "XZ", [[np.pi / 2 - dec_image, -ra_image + np.pi / 2]]
    ).as_matrix()[0]
    new_phase_direction_cosine = _directional_cosine(np.array([ra_image, dec_image]))

    uvw_rotmat = np.zeros((3, 3), np.double)
    phase_rotation = np.zeros((3,), np.double)

    rotmat_field_phase_direction = R.from_euler(
        "ZX",
        [
            [
                -np.pi / 2 + field_phase_direction.values[0],
                field_phase_direction.values[1] - np.pi / 2,
            ]
        ],
    ).as_matrix()[0]
    uvw_rotmat = np.matmul(rotmat_new_phase_direction, rotmat_field_phase_direction).T

    field_phase_direction_cosine = _directional_cosine(
        np.array(field_phase_direction.values)
    )
    phase_rotation = np.matmul(
        rotmat_new_phase_direction,
        (new_phase_direction_cosine - field_phase_direction_cosine),
    )

    return uvw_rotmat, phase_rotation


@jit(nopython=True, cache=True, nogil=True)
def _directional_cosine(phase_direction_in_radians):
    """
    # In https://arxiv.org/pdf/astro-ph/0207413.pdf see equation 160
    phase_direction_in_radians (RA,DEC)
    """

    phase_direction_cosine = np.zeros((3,), dtype=numba.f8)
    # phase_direction_cosine = np.zeros((3,))
    phase_direction_cosine[0] = np.cos(phase_direction_in_radians[0]) * np.cos(
        phase_direction_in_radians[1]
    )
    phase_direction_cosine[1] = np.sin(phase_direction_in_radians[0]) * np.cos(
        phase_direction_in_radians[1]
    )
    phase_direction_cosine[2] = np.sin(phase_direction_in_radians[1])
    return phase_direction_cosine


def _check_shift_parms(shift_parms):
    # from graphviper.parameter_checking.check_parms import check_parms
    from astroviper.utils.check_parms import check_parms, check_sel_parms
    import numbers

    parms_passed = True

    # if not(_check_parms(shift_parms, 'new_phase_direction', [dict])): parms_passed = False

    if not (
        check_parms(shift_parms, "common_tangent_reprojection", [bool], default=True)
    ):
        parms_passed = False

    if not (check_parms(shift_parms, "single_precision", [bool], default=True)):
        parms_passed = False

    return parms_passed



import numpy as np
from typing import List


def _fft_lm_to_uv(ary: np.ndarray, axes: List[int]) -> np.ndarray:
    """
    Do a 2-D FFT. This returns a complex-valued array
    :ary : np.ndarray
        Numpy array to FT
    :axes : List[int]
        List of two axes to transform
    Returns
    -------
    np.ndarray
    """
    return np.fft.fftshift(
        np.fft.fft2(np.fft.ifftshift(ary, axes=axes), axes=axes), axes=axes
    )



import numpy as np

# from graphviper.parameter_checking.check_parms import check_parms
from astroviper.utils.check_parms import check_parms


def _check_grid_parms(grid_parms):
    import numbers

    parms_passed = True
    arc_sec_to_rad = np.pi / (3600 * 180)

    if not (
        check_parms(
            grid_parms,
            "image_size",
            [list],
            list_acceptable_data_types=[int],
            list_len=2,
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            grid_parms,
            "image_center",
            [list],
            list_acceptable_data_types=[int],
            list_len=2,
            default=np.array(grid_parms["image_size"]) // 2,
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            grid_parms,
            "cell_size",
            [list],
            list_acceptable_data_types=[numbers.Number],
            list_len=2,
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            grid_parms,
            "fft_padding",
            [numbers.Number],
            default=1.2,
            acceptable_range=[1, 10],
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            grid_parms,
            "chan_mode",
            [str],
            acceptable_data=["cube", "continuum"],
            default="cube",
        )
    ):
        parms_passed = False

    if parms_passed == True:
        grid_parms["image_size"] = np.array(grid_parms["image_size"]).astype(int)
        grid_parms["image_size_padded"] = (
            grid_parms["fft_padding"] * grid_parms["image_size"]
        ).astype(int)
        grid_parms["image_center"] = np.array(grid_parms["image_center"])
        # grid_parms['cell_size'] = arc_sec_to_rad * np.array(grid_parms['cell_size'])
        # grid_parms['cell_size'][0] = -grid_parms['cell_size'][0]

    return parms_passed


##################### Function Specific Parms #####################
def _check_gcf_parms(gcf_parms):
    import numbers

    parms_passed = True

    if not (
        check_parms(
            gcf_parms,
            "function",
            [str],
            acceptable_data=["casa_airy", "airy"],
            default="casa_airy",
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            gcf_parms,
            "freq_chan",
            [list, np.array],
            list_acceptable_data_types=[numbers.Number],
            list_len=-1,
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            gcf_parms,
            "list_dish_diameters",
            [list, np.array],
            list_acceptable_data_types=[numbers.Number],
            list_len=-1,
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            gcf_parms,
            "list_blockage_diameters",
            [list, np.array],
            list_acceptable_data_types=[numbers.Number],
            list_len=-1,
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            gcf_parms,
            "unique_ant_indx",
            [list, np.array],
            list_acceptable_data_types=[numbers.Number],
            list_len=-1,
        )
    ):
        parms_passed = False

    if not (
        check_parms(
            gcf_parms,
            "pol",
            [list, np.array],
            list_acceptable_data_types=[numbers.Number, str],
            list_len=-1,
        )
    ):
        parms_passed = False
    if not (
        check_parms(gcf_parms, "chan_tolerance_factor", [numbers.Number], default=0.005)
    ):
        parms_passed = False
    if not (
        check_parms(
            gcf_parms,
            "oversampling",
            [list, np.array],
            list_acceptable_data_types=[int],
            list_len=2,
            default=[10, 10],
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            gcf_parms,
            "max_support",
            [list, np.array],
            list_acceptable_data_types=[int],
            list_len=2,
            default=[15, 15],
        )
    ):
        parms_passed = False
    # if not(check_parms(gcf_parms, 'image_phase_center', [list,np.array], list_acceptable_data_types=[numbers.Number], list_len=2)): parms_passed = False
    if not (
        check_parms(
            gcf_parms, "support_cut_level", [numbers.Number], default=2.5 * 10**-2
        )
    ):
        parms_passed = False
    if not (check_parms(gcf_parms, "a_chan_num_chunk", [int], default=3)):
        parms_passed = False

    if gcf_parms["function"] == "airy" or gcf_parms["function"] == "casa_airy":
        if not (
            check_parms(
                gcf_parms,
                "list_dish_diameters",
                [list, np.array],
                list_acceptable_data_types=[numbers.Number],
                list_len=-1,
            )
        ):
            parms_passed = False
        if not (
            check_parms(
                gcf_parms,
                "list_blockage_diameters",
                [list, np.array],
                list_acceptable_data_types=[numbers.Number],
                list_len=-1,
            )
        ):
            parms_passed = False

        if len(gcf_parms["list_dish_diameters"]) != len(
            gcf_parms["list_blockage_diameters"]
        ):
            print(
                "######### ERROR:Parameter ",
                "list_dish_diameters and list_blockage_diameters must be the same length.",
            )
            parms_passed = False

    if parms_passed == True:
        gcf_parms["oversampling"] = np.array(gcf_parms["oversampling"]).astype(int)
        gcf_parms["max_support"] = np.array(gcf_parms["max_support"]).astype(int)
        # gcf_parms['image_phase_center'] =  np.array(gcf_parms['image_phase_center'])
        gcf_parms["freq_chan"] = np.array(gcf_parms["freq_chan"])
        gcf_parms["list_dish_diameters"] = np.array(gcf_parms["list_dish_diameters"])
        gcf_parms["list_blockage_diameters"] = np.array(
            gcf_parms["list_blockage_diameters"]
        )
        gcf_parms["unique_ant_indx"] = np.array(gcf_parms["unique_ant_indx"])
        # gcf_parms['basline_ant'] =  np.array(gcf_parms['basline_ant'])
        gcf_parms["pol"] = np.array(gcf_parms["pol"])

    return parms_passed


def _check_mosaic_pb_parms(pb_mosaic_parms):
    parms_passed = True

    if not (check_parms(pb_mosaic_parms, "pb_name", [str], default="PB")):
        parms_passed = False

    if not (check_parms(pb_mosaic_parms, "weight_name", [str], default="WEIGHT_PB")):
        parms_passed = False

    return parms_passed


def _check_rotation_parms(rotation_parms):
    import numbers

    parms_passed = True

    if not (
        check_parms(
            rotation_parms,
            "new_phase_center",
            [list],
            list_acceptable_data_types=[numbers.Number],
            list_len=2,
        )
    ):
        parms_passed = False

    if not (
        check_parms(rotation_parms, "common_tangent_reprojection", [bool], default=True)
    ):
        parms_passed = False

    if not (check_parms(rotation_parms, "single_precision", [bool], default=True)):
        parms_passed = False

    return parms_passed


def _check_norm_parms(norm_parms):
    import numbers

    parms_passed = True

    if not (
        check_parms(
            norm_parms,
            "norm_type",
            [str],
            default="flat_sky",
            acceptable_data=["flat_noise", "flat_sky", "none"],
        )
    ):
        parms_passed = False

    if not (check_parms(norm_parms, "single_precision", [bool], default=True)):
        parms_passed = False

    if not (check_parms(norm_parms, "pb_limit", [numbers.Number], default=0.2)):
        parms_passed = False

    return parms_passed


def _check_imaging_weights_parms(imaging_weights_parms):
    import numbers

    parms_passed = True
    arc_sec_to_rad = np.pi / (3600 * 180)

    if not (
        check_parms(
            imaging_weights_parms,
            "weighting",
            [str],
            acceptable_data=["natural", "uniform", "briggs", "briggs_abs"],
            default="natural",
        )
    ):
        parms_passed = False

    if imaging_weights_parms["weighting"] == "briggs_abs":
        if not (
            check_parms(
                imaging_weights_parms, "briggs_abs_noise", [numbers.Number], default=1.0
            )
        ):
            parms_passed = False

    if (imaging_weights_parms["weighting"] == "briggs") or (
        imaging_weights_parms["weighting"] == "briggs_abs"
    ):
        if not (
            check_parms(
                imaging_weights_parms,
                "robust",
                [numbers.Number],
                default=0.5,
                acceptable_range=[-2, 2],
            )
        ):
            parms_passed = False

    return parms_passed


def _check_pb_parms(img_dataset, pb_parms):
    import numbers

    parms_passed = True
    arc_sec_to_rad = np.pi / (3600 * 180)

    # if not(check_parms(pb_parms, 'pb_name', [str], default='PB')): parms_passed = False

    if not (check_parms(pb_parms, "function", [str], default="casa_airy")):
        parms_passed = False

    if not (
        check_parms(
            pb_parms,
            "list_dish_diameters",
            [list],
            list_acceptable_data_types=[numbers.Number],
            list_len=-1,
        )
    ):
        parms_passed = False
    if not (
        check_parms(
            pb_parms,
            "list_blockage_diameters",
            [list],
            list_acceptable_data_types=[numbers.Number],
            list_len=-1,
        )
    ):
        parms_passed = False

    if len(pb_parms["list_dish_diameters"]) != len(pb_parms["list_blockage_diameters"]):
        print(
            "######### ERROR:Parameter ",
            "list_dish_diameters and list_blockage_diameters must be the same length.",
        )
        parms_passed = False

    return parms_passed



def _make_image_single_field(input_params):
    import time
    from xradio.correlated_data.load_processing_set import ProcessingSetIterator
    import toolviper.utils.logger as logger
    import dask

    # #print(input_params.keys())

    start_total = time.time()
    logger.debug("Processing chunk " + str(input_params["task_id"]))

    # dask.distributed.print('2 Input data ',input_params['task_id'], input_params['chunk_indices'], ' ***** ')

    start_0 = time.time()
    import numpy as np
    from astroviper._domain._imaging._make_imaging_weights import _make_imaging_weights
    from astroviper._domain._imaging._make_uv_sampling_grid import (
        _make_uv_sampling_grid_single_field,
    )
    from xradio.image import make_empty_sky_image
    from astroviper._domain._imaging._make_visibility_grid import (
        _make_visibility_grid_single_field,
    )
    from astroviper._domain._imaging._fft_norm_img_xds import _fft_norm_img_xds

    import xarray as xr

    logger.debug("0. Imports " + str(time.time() - start_0))

    start_1 = time.time()
    grid_params = input_params["grid_params"]

    image_freq_coord = input_params["task_coords"]["frequency"]["data"]

    if input_params["polarization"] is not None:
        image_polarization_coord = input_params["polarization"]
    else:
        image_polarization_coord = input_params["task_coords"]["polarization"]["data"]

    if input_params["time"] is not None:
        image_time_coord = input_params["time"]
    else:
        image_time_coord = input_params["task_coords"]["time"]["data"]

    logger.debug("Creating empty image ")
    img_xds = make_empty_sky_image(
        phase_center=grid_params["phase_direction"].values,
        image_size=grid_params["image_size"],
        cell_size=grid_params["cell_size"],
        chan_coords=image_freq_coord,
        pol_coords=image_polarization_coord,
        time_coords=image_time_coord,
    )
    img_xds.attrs["data_groups"] = {"mosaic": {}}

    T_empty_image = time.time() - start_1
    logger.debug("1. Empty Image " + str(time.time() - start_1))

    T_compute = 0.0
    T_load = 0.0
    T_weights = 0.0
    T_gcf = 0.0
    T_aperture_grid = 0.0
    T_uv_sampling_grid = 0.0
    T_vis_grid = 0.0

    data_group = input_params["data_group"]
    if data_group == "base":
        data_variables = ["FLAG", "UVW", "VISIBILITY", "WEIGHT"]
    elif data_group == "corrected":
        data_variables = ["FLAG", "UVW", "VISIBILITY_CORRECTED", "WEIGHT"]

    ps_iter = ProcessingSetIterator(
        input_params["data_selection"],
        input_params["input_data_store"],
        input_params["input_data"],
        data_variables=data_variables,
        load_sub_datasets=True,
    )
    logger.debug("1.5 Created ProcessingSetIterator ")

    from astroviper._domain._imaging._imaging_utils.gcf_prolate_spheroidal import (
        _create_prolate_spheroidal_kernel_1D,
    )

    cgk_1D = _create_prolate_spheroidal_kernel_1D(100, 7)

    start_2 = time.time()
    for ms_xds in ps_iter:
        start_compute = time.time()

        # Create a mask where baseline_antenna1_name does not equal baseline_antenna2_name
        mask = ms_xds["baseline_antenna1_name"] != ms_xds["baseline_antenna2_name"]
        # Apply the mask to the Dataset
        ms_xds = ms_xds.where(mask, drop=True)

        start_4 = time.time()
        data_group_out = _make_imaging_weights(
            ms_xds,
            grid_parms=grid_params,
            imaging_weights_parms={"weighting": "briggs", "robust": 0.6},
            sel_parms={"data_group_in": data_group},
        )
        T_weights = T_weights + time.time() - start_4

        start_7 = time.time()
        _make_uv_sampling_grid_single_field(
            ms_xds,
            cgk_1D,
            img_xds,
            vis_sel_parms={"data_group_in": data_group_out},
            img_sel_parms={"data_group_in": "mosaic"},
            grid_parms=grid_params,
        )  # Will become the PSF.
        T_uv_sampling_grid = T_uv_sampling_grid + time.time() - start_7

        start_8 = time.time()
        _make_visibility_grid_single_field(
            ms_xds,
            cgk_1D,
            img_xds,
            vis_sel_parms={"data_group_in": data_group_out},
            img_sel_parms={"data_group_in": "mosaic"},
            grid_parms=grid_params,
        )
        T_vis_grid = T_vis_grid + time.time() - start_8
        T_compute = T_compute + time.time() - start_compute

    # print(img_xds)
    from astroviper._domain._imaging._imaging_utils._make_pb_symmetric import (
        _airy_disk_rorder,
    )

    pb_parms = {}
    pb_parms["list_dish_diameters"] = np.array([10.7])
    pb_parms["list_blockage_diameters"] = np.array([0.75])
    pb_parms["ipower"] = 1

    grid_params["image_center"] = (np.array(grid_params["image_size"]) // 2).tolist()
    # (1, 1, len(pol), 1, 1))
    # print(_airy_disk_rorder(ms_xds.frequency.values, ms_xds.polarization.values, pb_parms, grid_params).shape)

    # img_xds["PRIMARY_BEAM"] =  xr.DataArray(_airy_disk_rorder(ms_xds.frequency.values, ms_xds.polarization.values, pb_parms, grid_params)[0,...], dims=("frequency", "polarization", "l", "m"))
    img_xds["PRIMARY_BEAM"] = xr.DataArray(
        np.ones(img_xds.UV_SAMPLING.shape), dims=("frequency", "polarization", "l", "m")
    )

    T_load = time.time() - start_2 - T_compute

    logger.debug("2. Load " + str(T_load))
    logger.debug("3. Weights " + str(T_weights))
    logger.debug("5. make_gridding_convolution_function " + str(T_gcf))
    logger.debug("6. Aperture grid " + str(T_aperture_grid))
    logger.debug("7. UV sampling grid " + str(T_uv_sampling_grid))
    logger.debug("8. Visibility grid " + str(T_vis_grid))
    logger.debug("Compute " + str(T_compute))

    start_9 = time.time()

    gcf_xds = xr.Dataset()
    gcf_xds.attrs["oversampling"] = [100, 100]
    gcf_xds.attrs["SUPPORT"] = [7, 7]
    from astroviper._domain._imaging._imaging_utils.gcf_prolate_spheroidal import (
        _create_prolate_spheroidal_kernel,
    )

    _, ps_corr_image = _create_prolate_spheroidal_kernel(
        100, 7, n_uv=img_xds["UV_SAMPLING"].shape[-2:]
    )

    # print(ps_corr_image.shape)
    gcf_xds["PS_CORR_IMAGE"] = xr.DataArray(ps_corr_image, dims=("l", "m"))

    _fft_norm_img_xds(
        img_xds,
        gcf_xds,
        grid_params,
        norm_parms={},
        sel_parms={"data_group_in": "mosaic", "data_group_out": "mosaic"},
    )
    T_fft = time.time() - start_9
    logger.debug("9. fft norm " + str(time.time() - start_9))

    # # Tranform uv-space -> lm-space (sky)

    start_10 = time.time()
    parallel_dims_chunk_id = dict(
        zip(input_params["parallel_dims"], input_params["chunk_indices"])
    )

    from xradio.image._util._zarr.zarr_low_level import write_chunk
    import os

    # dask.distributed.print('Writing results to Lustre for task ',input_params['task_id'],' ***** ')
    img_xds = img_xds.transpose("polarization", "frequency", ...).expand_dims(
        dim="dummy", axis=0
    )

    if input_params["to_disk"]:
        for data_variable, meta in input_params["zarr_meta"].items():
            write_chunk(
                img_xds,
                meta,
                parallel_dims_chunk_id,
                input_params["compressor"],
                input_params["image_file"],
            )
    else:
        return img_xds

    # mini_cube_name = os.path.join(input_params["image_file"], 'cube_chunk_' + str(input_params["task_id"]))
    # logger.debug('The mini_cube name ' + mini_cube_name)
    # img_xds.attrs.clear()
    # with dask.config.set(scheduler="synchronous"):
    #     img_xds.to_zarr(mini_cube_name)

    T_to_disk = time.time() - start_10
    logger.debug("10. to disk " + str(time.time() - start_10))
    # dask.distributed.print('Done writing results to Lustre for task ',input_params['task_id'], T_to_disk,' ***** ')

    logger.debug(
        "Completed task "
        + str(input_params["task_id"])
        + " in "
        + str(time.time() - start_total)
        + ", "
        + "Empty_image, Load, Weights, GCF, Aperture, UV, Vis, Total Loop, FFT, To_disk "
        + str(T_empty_image)
        + ", "
        + str(T_load)
        + ", "
        + str(T_weights)
        + ", "
        + str(T_gcf)
        + ", "
        + str(T_aperture_grid)
        + ", "
        + str(T_uv_sampling_grid)
        + ", "
        + str(T_vis_grid)
        + ", "
        + str(T_compute)
        + ", "
        + str(T_fft)
        + ", "
        + str(T_to_disk)
    )
    # logger.debug("Completed task " + str(input_params['task_id']) + " in " + str(time.time()-start_total) + " s.")
    # logger.debug("***"*20)
    import pandas as pd

    return_dict = {
        "task_id": [input_params["task_id"]],
        "n_channels": [len(input_params["task_coords"]["frequency"]["data"])],
        "T_load": T_load,
        "T_empty_image": T_empty_image,
        "T_load": T_load,
        "T_weights": T_weights,
        "T_aperture_grid": T_aperture_grid,
        "T_uv_sampling_grid": T_uv_sampling_grid,
        "T_vis_grid": T_vis_grid,
        "T_compute": T_compute,
        "T_fft": T_fft,
        "T_to_disk": T_to_disk,
    }
    df = pd.DataFrame(return_dict)

    # import pandas as pd
    # df = pd.DataFrame()

    return df



#!/usr/bin/env python
# coding: utf-8

from astropy import units as u
from astroviper._domain._imaging._fft import _fft_lm_to_uv
from astroviper._domain._imaging._ifft import _ifft_uv_to_lm
import copy
import dask
import dask.array as da
from graphviper.graph_tools.coordinate_utils import (
    interpolate_data_coords_onto_parallel_coords,
)
from graphviper.graph_tools.coordinate_utils import make_parallel_coord
from graphviper.graph_tools.map import map
from toolviper.utils.display import dict_to_html
import numpy as np
import os
import shutil
import time
from typing import Union
import xarray as xr
from xradio.image import read_image
from xradio.image import make_empty_aperture_image
from xradio.image import load_image
from xradio.image import write_image
import toolviper.utils.logger as logger

_sky = "SKY"
_beam = "BEAM"


def _compute_u_v(xds):
    shape = [xds.dims["l"], xds.dims["m"]]
    # sics = np.abs(xds.attrs["direction"]["reference"]["cdelt"])
    sics = np.abs(2 * [xds.l[1] - xds.l[0]])
    w_xds = make_empty_aperture_image(
        phase_center=[0, 0],
        image_size=shape,
        sky_image_cell_size=sics,
        chan_coords=[1],
        pol_coords=["I"],
        time_coords=[0],
    )
    u = np.zeros(shape)
    v = np.zeros(shape)
    for i, ux in enumerate(w_xds.coords["u"]):
        u[i, :] = ux
    for i, vx in enumerate(w_xds.coords["v"]):
        v[:, i] = vx
    return (u, v)


def _feather(input_params):
    # display(HTML(dict_to_html(input_params)))

    def _compute_w_multiple_beams(xds, uv):
        """xds is the single dish xds"""
        beams = xds[_beam]
        # logger.debug("beams " + str(beams))
        w = np.zeros(xds[_sky].shape)
        bunit = beams.attrs["units"]
        bmaj = beams.sel(beam_param="major")
        # logger.debug("bmaj orig shape" + str(bmaj.shape))
        # add l and m dims
        bmaj = np.expand_dims(bmaj, -1)
        bmaj = np.expand_dims(bmaj, -1)
        # logger.debug("bmaj shape " + str(bmaj.shape))
        alpha = bmaj * u.Unit(bunit)
        alpha = alpha.to(u.rad).value
        bmin = beams.sel(beam_param="minor")
        bmin = np.expand_dims(bmin, -1)
        bmin = np.expand_dims(bmin, -1)
        beta = bmin * u.Unit(bunit)
        beta = beta.to(u.rad).value
        bpa = beams.sel(beam_param="pa")
        bpa = np.expand_dims(bpa, -1)
        bpa = np.expand_dims(bpa, -1)
        phi = bpa * u.Unit(bunit)
        phi = phi.to(u.rad).value

        alpha2 = alpha * alpha
        beta2 = beta * beta
        # u -> uu, v -> vv because we've already used
        # u for astropy.units
        uu, vv = uv
        uu = uu[np.newaxis, np.newaxis, np.newaxis, :, :]
        vv = vv[np.newaxis, np.newaxis, np.newaxis, :, :]
        aterm2 = (uu * np.sin(phi) - vv * np.cos(phi)) ** 2
        bterm2 = (uu * np.cos(phi) + vv * np.sin(phi)) ** 2
        w = np.exp(
            -np.pi * np.pi / 4.0 / np.log(2) * (alpha2 * aterm2 + beta2 * bterm2)
        )
        # w is an np.array
        return w

    # if input_params["input_data"] is None: #Load
    dtypes = {"sd": np.int32, "int": np.int32}
    for k in ["sd", "int"]:
        # the "data_selection" key is set in
        # interpolate_data_coords_onto_parallel_coords()
        # print("data store", input_params["input_data_store"][k])
        # print("block_des", input_params["data_selection"][k])
        xds = load_image(
            input_params["input_data_store"][k],
            block_des=input_params["data_selection"][k],
        )
        # print("load image for", k, "complete")
        # print("completed load_image()")
        fft_plane = (
            xds[_sky].dims.index(input_params["axes"][0]),
            xds[_sky].dims.index(input_params["axes"][1]),
        )
        # print("completed fft_plane")
        # else:
        #   img_xds = input_params["input_data"]['img'] #In memory
        aperture = _fft_lm_to_uv(xds[_sky], fft_plane)
        # print("completed _fft_im_to_uv()")
        dtypes[k] = xds[_sky].dtype
        if k == "int":
            int_ap = aperture
            int_xds = xds
            # logger.debug("int_xds beam " + str(int_xds[_beam]))
        else:
            sd_ap = aperture
            sd_xds = xds
            # logger.debug("sd_xds beam " + str(sd_xds[_beam]))
    mytype = dtypes["sd"] if dtypes["sd"] < dtypes["int"] else dtypes["int"]

    uv = _compute_u_v(sd_xds)
    w = _compute_w_multiple_beams(sd_xds, uv)

    one_minus_w = 1 - w
    s = input_params["s"]
    if _beam in int_xds.data_vars:
        int_ba = int_xds[_beam].sel(beam_param="major") * int_xds[_beam].sel(
            beam_param="minor"
        )
    else:
        error_message = "Unable to find BEAM data variable in interferometer image."
        logger.error(error_message)
        raise Exception(error_message)

    if _beam in sd_xds.data_vars:
        sd_ba = sd_xds[_beam].sel(beam_param="major") * sd_xds[_beam].sel(
            beam_param="minor"
        )
    else:
        error_message = "Unable to find BEAM data variable in single dish image."
        logger.error(error_message)
        raise Exception(error_message)
    # need to use values becuase the obs times will in general be different
    # which would cause a resulting shape with the time dimension having
    # length 0
    beam_ratio_values = int_ba.values / sd_ba.values
    # use interferometer coords
    beam_ratio = xr.DataArray(
        beam_ratio_values, dims=int_ba.dims, coords=int_ba.coords.copy()
    )

    beam_ratio = np.expand_dims(beam_ratio, -1)
    beam_ratio = np.expand_dims(beam_ratio, -1)

    term = (one_minus_w * int_ap + s * beam_ratio * sd_ap) / (one_minus_w + s * w)
    feather_npary = _ifft_uv_to_lm(term, fft_plane).astype(mytype)
    from xradio.image._util._zarr.zarr_low_level import write_chunk

    from xradio.image import make_empty_sky_image

    """
    # FIXME lon/latpole is not the phase center
    phase_center = [
        sd_xds.attrs["direction"]["longpole"]["value"],
        int_xds.attrs["direction"]["latpole"]["value"],
    ]

    featherd_img_chunk_xds = make_empty_sky_image(
        phase_center=phase_center,
        image_size=[int_xds.sizes["l"], int_xds.sizes["m"]],
        cell_size=int_xds.attrs["direction"]["reference"]["cdelt"],
        chan_coords=int_xds.frequency.values,
        pol_coords=int_xds.polarization.values,
        time_coords=[0],
    )
    """

    featherd_img_chunk_xds = xr.Dataset(coords=int_xds.coords)
    # we need an xradio function to return an ordered list of dimensions
    featherd_img_chunk_xds[_sky] = xr.DataArray(
        feather_npary, dims=["time", "frequency", "polarization", "l", "m"]
    )
    parallel_dims_chunk_id = dict(
        zip(input_params["parallel_dims"], input_params["chunk_indices"])
    )

    # print('input_params["zarr_meta"]',input_params["zarr_meta"])
    if input_params["to_disk"]:
        for data_variable, meta in input_params["zarr_meta"].items():
            write_chunk(
                featherd_img_chunk_xds,
                meta,
                parallel_dims_chunk_id,
                input_params["compressor"],
                input_params["image_file"],
            )

        results_dict = {}
        return results_dict
    else:
        results_dict = {"featherd_img_chunk_xds": featherd_img_chunk_xds}
        return featherd_img_chunk_xds


from numcodecs import Blosc


def feather(
    outim: Union[dict, None],
    highres: str,
    lowres: str,
    sdfactor: float,
    selection: dict = {},
    thread_info=None,
    compressor=Blosc(cname="lz4", clevel=5),
):
    """
    Create an image from a single dish and interferometer image using the
    feather algorithm
    Parameters
    ----------
    outim : output image information, dict or None
        if None, no image is written (probably only useful for debugging). if
        a dict it must have a "name" key. "name" is the directory to which the
        zarr format image is written. An "overwrite" boolean parameter is
        optional. If it does not exist, it is assumed that the user does not
        want to overwrite an already extant image of the same name. Note that
        feather only writes zarr format images. If another output format is
        desired, the user/caller must convert the zarr format image to the
        desired format after running this function. The zarr file is used to
        accumulate the results of the computation, chunk by chunk,and so it is
        needed at the start of the computation and is not written in total at
        the end but rather is written to as each chunk is computed.
    highres : interferometer image, string or xr.Dataset.
        If str, an image file by that name will be read from disk.
        If xr.Dataset, that xds will be used for the interferometer image.
    lowres : Single dish image, string or xr.Dataset.
        If str, an image file by that name will be read from disk.
        If xr.Dataset, that xds will be used for the single dish image.
    """
    if outim is not None:
        if type(outim) != dict:
            raise ValueError(
                "If specified, outim must be a dictionary with required key "
                "'name' and optional key 'overwrite'."
            )
        if "name" not in outim:
            raise ValueError("If specfied, outim dict must have key 'name'.")
        if "overwrite" not in outim:
            outim["overwrite"] = False
        elif type(outim["overwrite"]) != bool:
            raise TypeError("If specified, outim['overwrite'] must be a boolean value")
        if not outim["overwrite"]:
            if os.path.exists(outim["name"]):
                raise RuntimeError(
                    f"Already existing file {outim['name']} will not be "
                    "overwritten. To overwrite it, set outim['overwrite'] = True"
                )

    # Read in input images
    # single dish image
    sd_xds = (
        read_image(lowres, selection=selection) if isinstance(lowres, str) else lowres
    )
    # interferometer image
    int_xds = (
        read_image(highres, selection=selection)
        if isinstance(highres, str)
        else highres
    )
    if sd_xds[_sky].shape != int_xds[_sky].shape:
        raise RuntimeError("Image shapes differ")

    # Determine chunking
    from astroviper._utils.data_partitioning import bytes_in_dtype

    ## Determine the amount of memory required by the node task if all dimensions that chunking will occur on are singleton.
    ## For example feather does chunking only only frequency, so memory_singleton_chunk should be the amount of memory requered by _feather when there is a single frequency channel.
    singleton_chunk_sizes = dict(sd_xds[_sky].sizes)
    del singleton_chunk_sizes[
        "frequency"
    ]  # Remove dimensions that will be chuncked on.
    fudge_factor = 1.1
    n_images_in_memory = 3.0  # Two input and one output image.
    memory_singleton_chunk = (
        n_images_in_memory
        * np.prod(np.array(list(singleton_chunk_sizes.values())))
        * fudge_factor
        * bytes_in_dtype[str(sd_xds[_sky].dtype)]
        / (1024**3)
    )

    chunking_dims_sizes = {
        "frequency": int_xds[_sky].sizes["frequency"]
    }  # Need to know how many frequency channels there are.
    from astroviper._utils.data_partitioning import (
        calculate_data_chunking,
        get_thread_info,
    )

    if thread_info is None:
        thread_info = get_thread_info()
    logger.debug("Thread info " + str(thread_info))
    n_chunks_dict = calculate_data_chunking(
        memory_singleton_chunk,
        chunking_dims_sizes,
        thread_info,
        constant_memory=0,
        tasks_per_thread=4,
    )

    parallel_coords = {}
    parallel_coords["frequency"] = make_parallel_coord(
        coord=sd_xds.frequency, n_chunks=n_chunks_dict["frequency"]
    )
    # display(HTML(dict_to_html(parallel_coords["frequency"])))

    # FIXME need to do this for int_xds as well, can I do in one go
    # by making "img" a list of xdses?
    # JW says multiple items in the dict, the key names aren't important so can
    # be anything contextual to the problem at hand
    input_data = {"sd": sd_xds, "int": int_xds}
    node_task_data_mapping = interpolate_data_coords_onto_parallel_coords(
        parallel_coords, input_data
    )
    # display(HTML(dict_to_html(node_task_data_mapping)))

    # Create empty image on disk
    to_disk = True
    if to_disk:
        # create new xarray.Dataset with coordinates same as int_xds
        # but with no data

        """
        from xradio.image import make_empty_sky_image

        # the phase center is not the same as lon/lat pole
        phase_center = [
            sd_xds.attrs["direction"]["lonpole"]["data"],
            int_xds.attrs["direction"]["latpole"]["data"],
        ]

        featherd_img_xds = make_empty_sky_image(
            phase_center=phase_center,
            image_size=[int_xds.sizes["l"], int_xds.sizes["m"]],
            cell_size=int_xds.attrs["direction"]["reference"]["cdelt"],
            chan_coords=parallel_coords["frequency"]["data"],
            pol_coords=int_xds.polarization.values,
            time_coords=[0],
        )
        """

        featherd_img_xds = xr.Dataset(coords=int_xds.coords)
        featherd_img_xds.attrs = copy.deepcopy(int_xds.attrs)
        # we cannot build the beam in parallel because its parallel dims are no l, m
        # so just copy the whole thing here
        featherd_img_xds[_beam] = int_xds[_beam].copy()

        write_image(
            featherd_img_xds,
            imagename=outim["name"],
            out_format="zarr",
            overwrite=outim["overwrite"],
        )

        # Create the empty data variable SKY.
        from xradio.image._util._zarr.zarr_low_level import (
            create_data_variable_meta_data,
        )

        if int_xds[_sky].dtype == np.float32:
            from xradio.image._util._zarr.zarr_low_level import (
                image_data_variables_and_dims_single_precision as image_data_variables_and_dims,
            )
        elif int_xds[_sky].dtype == np.float64:
            from xradio.image._util._zarr.zarr_low_level import (
                image_data_variables_and_dims_double_precision as image_data_variables_and_dims,
            )
        else:
            error_message = (
                "Unsupported data type of image "
                + str(int_xds[_sky].dtype)
                + " expected float32 or float64."
            )
            logger.error(error_message)
            raise Exception(error_message)

        xds_dims = dict(int_xds.sizes)
        # right now the keys are lower case, but the associated values are all caps
        # the beam cannot be written in chunks because
        # ValueError: could not broadcast input array from shape (1,4,1,3) into shape (1,4,1,1024,1024)

        data_variables = ["sky"]
        data_varaibles_and_dims_sel = {
            key: image_data_variables_and_dims[key] for key in data_variables
        }
        zarr_meta = create_data_variable_meta_data(
            outim["name"],
            data_varaibles_and_dims_sel,
            xds_dims,
            parallel_coords,
            compressor,
        )

    input_params = {}
    input_params["input_data_store"] = {"sd": lowres, "int": highres}
    input_params["axes"] = ("l", "m")  # (3,4)
    input_params["image_file"] = outim["name"]
    input_params["s"] = 1

    if to_disk:
        input_params["to_disk"] = to_disk
        input_params["compressor"] = compressor
        input_params["zarr_meta"] = zarr_meta

    t0 = time.time()
    viper_graph = map(
        input_data=input_data,
        node_task_data_mapping=node_task_data_mapping,
        node_task=_feather,
        input_params=input_params,
        in_memory_compute=False,
    )
    from graphviper.graph_tools import generate_dask_workflow

    dask_graph = generate_dask_workflow(viper_graph)
    logger.debug("Time to create graph " + str(time.time() - t0))

    # dask.visualize(graph, filename="map_graph")
    t0 = time.time()
    res = dask.compute(dask_graph)
    logger.info("Time to compute() feather " + str(time.time() - t0) + "s")

    import zarr

    zarr.consolidate_metadata(outim["name"])



import numpy as np
import scipy
from scipy import constants
from numba import jit
import numba
import xarray as xr

# from graphviper.parameter_checking.check_parms import check_sel_parms
from astroviper.utils.check_parms import check_parms, check_sel_parms
from astroviper._domain._imaging._check_imaging_parameters import (
    _check_grid_parms,
    _check_norm_parms,
)
import copy


def _fft_norm_img_xds(img_xds, gcf_xds, grid_parms, norm_parms, sel_parms):
    _sel_parms = copy.deepcopy(sel_parms)
    _grid_parms = copy.deepcopy(grid_parms)
    assert _check_grid_parms(_grid_parms), "######### ERROR: grid_parms checking failed"

    _norm_parms = copy.deepcopy(norm_parms)
    assert _check_norm_parms(_norm_parms), "######### ERROR: norm_parms checking failed"

    if "PRIMARY_BEAM" in img_xds:
        data_group_in, data_group_out = check_sel_parms(
            img_xds,
            _sel_parms,
            default_data_group_out={
                "imaging": {
                    "point_spread_function": "POINT_SPREAD_FUNCTION",
                    "sky": "SKY",
                }
            },
        )
        data_group_out["primary_beam"] = "PRIMARY_BEAM"
    else:
        data_group_in, data_group_out = check_sel_parms(
            img_xds,
            _sel_parms,
            default_data_group_out={
                "imaging": {
                    "primary_beam": "PRIMARY_BEAM",
                    "point_spread_function": "POINT_SPREAD_FUNCTION",
                    "sky": "SKY",
                }
            },
        )
    fft_pair = {
        "aperture": "primary_beam",
        "uv_sampling": "point_spread_function",
        "visibility": "sky",
    }
    sum_of_weight_pair = {
        "aperture": "aperture_grid_normalization",
        "uv_sampling": "uv_sampling_normalization",
        "visibility": "visibility_normalization",
    }

    # print(data_group_in, data_group_out)

    for data_variable in ["aperture", "uv_sampling", "visibility"]:
        if data_variable in data_group_in.keys():

            if (data_variable == "aperture") and ("APERTURE" in img_xds):
                # print('1.',data_group_out["aperture_grid_sum_weight"],img_xds[data_group_in[data_variable]].shape,_grid_parms['image_size'])
                image = _remove_padding(
                    _ifft_uv_to_lm(img_xds[data_group_in[data_variable]].values),
                    _grid_parms["image_size"],
                ).real
                sum_weight_copy = copy.deepcopy(
                    img_xds[data_group_out["aperture_normalization"]].values
                )
                # print("%%%%%%%%%%%%%Sum of weight", sum_weight_copy,data_group_out[sum_of_weight_pair[data_variable]])
                ##Don't mutate inputs, therefore do deep copy (https://docs.dask.org/en/latest/delayed-best-practices.html).
                sum_weight_copy[sum_weight_copy == 0] = 1
                image = np.sqrt(np.abs(image / sum_weight_copy[:, :, None, None]))
                if _norm_parms["pb_limit"] > 0:
                    image[image < _norm_parms["pb_limit"]] = np.nan  # 0.0
            else:
                image = _remove_padding(
                    _ifft_uv_to_lm(img_xds[data_group_in[data_variable]].values),
                    _grid_parms["image_size"],
                )
                sum_weight = img_xds[
                    data_group_out[sum_of_weight_pair[data_variable]]
                ].values
                # print("%%%%%%%%%%%%%Sum of weight", sum_weight,data_group_out[sum_of_weight_pair[data_variable]])
                primary_beam = img_xds[data_group_out["primary_beam"]].values
                oversampling = gcf_xds.oversampling
                ps_correcting_image = gcf_xds.PS_CORR_IMAGE.values
                direction = "forward"

                if data_variable == "uv_sampling":
                    norm_parms["norm_type"] = "flat_noise"
                if data_variable == "visibility":
                    norm_parms["norm_type"] = "flat_sky"
                image = _normalize(
                    image,
                    sum_weight,
                    primary_beam,
                    oversampling,
                    ps_correcting_image,
                    direction,
                    norm_parms=_norm_parms,
                )
            img_xds[data_group_out[fft_pair[data_variable]]] = xr.DataArray(
                image, dims=("frequency", "polarization", "l", "m")
            )


def _ifft_uv_to_lm(image):
    image_shape = image.shape
    return np.fft.fftshift(
        np.fft.ifft2(np.fft.ifftshift(image, axes=(2, 3)), axes=(2, 3)), axes=(2, 3)
    ).real * (image_shape[-1] * image_shape[-2])

    # dafft.fftshift(dafft.ifft2(dafft.ifftshift(grids_and_sum_weights[0], axes=(0, 1)), axes=(0, 1)), axes=(0, 1))


def _fft_lm_to_uv(image):
    return np.fft.fftshift(
        np.fft.fft2(np.fft.ifftshift(image, axes=(2, 3)), axes=(2, 3)), axes=(2, 3)
    ).real


def _remove_padding(image_dask_array, image_size):
    # Check that image_size < image_size
    # Check parameters

    import numpy as np

    image_size_padded = np.array(image_dask_array.shape[-2:])
    start_xy = image_size_padded // 2 - image_size // 2
    end_xy = start_xy + image_size

    # print("start end",image_size_padded,start_xy,end_xy)

    image_dask_array = image_dask_array[
        ..., start_xy[0] : end_xy[0], start_xy[1] : end_xy[1]
    ]
    return image_dask_array


def _normalize(
    image,
    sum_weight,
    primary_beam,
    oversampling,
    ps_correcting_image,
    direction,
    norm_parms,
):
    """
    PB normalization on the cubes

    direction : 'forward''reverse'
    norm_type : 'flatnoise','flatsky','common','pbsquare'

    Multiply and/or divide by PB models, accounting for masks/regions.

    #See https://casa.nrao.edu/casadocs/casa-5.6.0/imaging/synthesis-imaging/data-weighting Normalizationm Steps
    #Standard gridder (only ps term) devide by sum of weight and ps correcting image.
    #https://library.nrao.edu/public/memos/evla/EVLAM_198.pdf

    #

    """
    import dask.array as da
    import numpy as np
    import copy

    norm_type = norm_parms["norm_type"]

    def normalize_image(
        image, sum_weights, normalizing_image, oversampling, correct_oversampling
    ):
        sum_weights_copy = copy.deepcopy(
            sum_weights
        )  ##Don't mutate inputs, therefore do deep copy (https://docs.dask.org/en/latest/delayed-best-practices.html).
        sum_weights_copy[sum_weights_copy == 0] = 1

        if correct_oversampling:
            image_size = np.array(image.shape[2:])
            image_center = image_size // 2
            sincx = np.sinc(
                np.arange(-image_center[0], image_size[0] - image_center[0])
                / (image_size[0] * oversampling[0])
            )
            sincy = np.sinc(
                np.arange(-image_center[1], image_size[1] - image_center[1])
                / (image_size[1] * oversampling[1])
            )

            oversampling_correcting_func = np.dot(
                sincx[:, None], sincy[None, :]
            )  # Last section for sinc correcting function https://library.nrao.edu/public/memos/evla/EVLAM_198.pdf

            # print(image.shape,sum_weights_copy.shape,oversampling_correcting_func.shape,oversampling_correcting_func[None,None,:,:].shape,normalizing_image.shape)

            normalized_image = (image / sum_weights_copy) / (
                oversampling_correcting_func[None, None, :, :] * normalizing_image
            )

            # print(sum_weights_copy,oversampling_correcting_func[500,360,None,None])
            # normalized_image = (image / sum_weights_copy ) / (oversampling_correcting_func[:,:,None,None])
        else:
            normalized_image = (image / sum_weights_copy) / normalizing_image
        return normalized_image

    if direction == "forward":
        correct_oversampling = True
        if norm_type == "flat_noise":
            # Divide the raw image by sqrt(.weight) so that the input to the minor cycle represents the product of the sky and PB. The noise is 'flat' across the region covered by each PB.
            normalizing_image = ps_correcting_image[None, None, :, :] * primary_beam
            normalized_image = normalize_image(
                image,
                sum_weight[:, :, None, None],
                normalizing_image,
                oversampling,
                correct_oversampling,
            )
        elif norm_type == "flat_sky":
            #  Divide the raw image by .weight so that the input to the minor cycle represents only the sky. The noise is higher in the outer regions of the primary beam where the sensitivity is low.
            normalizing_image = (
                ps_correcting_image[None, None, :, :] * primary_beam * primary_beam
            )

            # print(sel_parms['data_group_in']['weight_pb'])
            # print('$%$%',img_dataset[sel_parms['data_group_in']['weight_pb']].data.compute())

            normalized_image = normalize_image(
                image,
                sum_weight[:, :, None, None],
                normalizing_image,
                oversampling,
                correct_oversampling,
            )

            # print(normalized_image.compute())
        elif norm_type == "none":
            pass
            # print('in normalize none ')
            # No normalization after gridding and FFT. The minor cycle sees the sky times pb square
        #            normalizing_image = ps_correcting_image[None, None, :, :]
        #            normalized_image = da.map_blocks(
        #                normalize_image,
        #                image,
        #                sum_weight[:, :, None, None],
        #                normalizing_image,
        #                oversampling,
        #                correct_oversampling,
        #            )
        # normalized_image = image

        if norm_parms["pb_limit"] > 0:
            normalized_image[primary_beam < norm_parms["pb_limit"]] = np.nan  # 0.0

        if norm_parms["single_precision"]:
            normalized_image = (normalized_image.astype(np.float32)).astype(np.float64)

        return normalized_image
    elif direction == "reverse":
        print("reverse operation not yet implemented not yet implemented")



from ._fft import _fft_lm_to_uv
from ._ifft import _ifft_uv_to_lm
from ._make_imaging_weights import _make_imaging_weights
from ._make_gridding_convolution_function import _make_gridding_convolution_function
from ._make_aperture_grid import _make_aperture_grid
from ._make_uv_sampling_grid import _make_uv_sampling_grid
from ._make_visibility_grid import _make_visibility_grid



import numpy as np
import scipy

# import cngi._utils._constants as const
from scipy import constants
from numba import jit
import numba
import xarray as xr
from astroviper._domain._imaging._check_imaging_parameters import (
    _check_imaging_weights_parms,
    _check_grid_parms,
)
from astroviper._domain._imaging._imaging_utils._standard_grid import (
    _standard_imaging_weight_degrid_numpy_wrap,
    _standard_grid_psf_numpy_wrap,
)

# from graphviper.parameter_checking.check_parms import check_sel_parms
from astroviper.utils.check_parms import check_parms, check_sel_parms
import copy


def _make_imaging_weights(ms_xds, grid_parms, imaging_weights_parms, sel_parms):
    _sel_parms = copy.deepcopy(sel_parms)
    _imaging_weights_parms = copy.deepcopy(imaging_weights_parms)
    assert _check_imaging_weights_parms(
        _imaging_weights_parms
    ), "######### ERROR: imaging_weights_parms checking failed"

    if _imaging_weights_parms["weighting"] == "natural":
        _sel_parms["overwrite"] = True  # No actual overwrite is occuring.
        data_group_in, data_group_out = check_sel_parms(
            ms_xds,
            _sel_parms,
            default_data_group_out={"imaging": {"weight_imaging": "WEIGHT"}},
        )
        return _sel_parms["data_group_out"]
    else:
        data_group_in, data_group_out = check_sel_parms(
            ms_xds,
            _sel_parms,
            default_data_group_out={"imaging": {"weight_imaging": "WEIGHT_IMAGING"}},
        )

    # print(data_group_in, data_group_out)

    _grid_parms = copy.deepcopy(grid_parms)
    assert _check_grid_parms(_grid_parms), "######### ERROR: grid_parms checking failed"
    _grid_parms["image_size_padded"] = _grid_parms[
        "image_size"
    ]  # do not need to pad since no fft
    _grid_parms["oversampling"] = 0
    _grid_parms["support"] = 1
    _grid_parms["do_psf"] = True
    _grid_parms["complex_grid"] = False
    _grid_parms["do_imaging_weight"] = True

    cgk_1D = np.ones((1))

    uvw = ms_xds[data_group_out["uvw"]].values
    data_weight = ms_xds[data_group_out["weight"]].values
    freq_chan = ms_xds.frequency.values

    # Grid Weights
    weight_density_grid, sum_weight = _standard_grid_psf_numpy_wrap(
        uvw, data_weight, freq_chan, cgk_1D, _grid_parms
    )

    # Calculate Briggs
    # print('weight_density_grid',weight_density_grid)
    briggs_factors = _calculate_briggs_parms(
        weight_density_grid, sum_weight, _imaging_weights_parms
    )  # 2 x chan x pol

    imaging_weights = _standard_imaging_weight_degrid_numpy_wrap(
        weight_density_grid, uvw, data_weight, briggs_factors, freq_chan, _grid_parms
    )

    ms_xds[data_group_out["weight_imaging"]] = xr.DataArray(
        imaging_weights, dims=ms_xds[data_group_out["weight"]].dims
    )

    return _sel_parms["data_group_out"]


def _calculate_briggs_parms(grid_of_imaging_weights, sum_weight, imaging_weights_parms):
    if imaging_weights_parms["weighting"] == "briggs":
        robust = imaging_weights_parms["robust"]
        briggs_factors = np.ones((2,) + sum_weight.shape)

        squared_sum_weight = np.sum((grid_of_imaging_weights) ** 2, axis=(2, 3))
        briggs_factors[0, :, :] = (
            np.square(5.0 * 10.0 ** (-robust)) / (squared_sum_weight / sum_weight)
        )[None, None, :, :]
    elif imaging_weights_parms["weighting"] == "briggs_abs":
        robust = imaging_weights_parms["robust"]
        briggs_factors = np.ones((2,) + sum_weight.shape)
        briggs_factors[0, :, :] = briggs_factor[0, 0, 0, :, :] * np.square(robust)
        briggs_factors[1, :, :] = (
            briggs_factor[1, 0, 0, :, :]
            * 2.0
            * np.square(imaging_weights_parms["briggs_abs_noise"])
        )
    else:
        briggs_factors = np.zeros((2, 1, 1) + sum_weight.shape)
        briggs_factors[0, :, :] = np.ones((1, 1, 1) + sum_weight.shape)

    return briggs_factors



import numpy as np


def check_parms(
    parm_dict,
    string_key,
    acceptable_data_types,
    acceptable_data=None,
    acceptable_range=None,
    list_acceptable_data_types=None,
    list_len=None,
    default=None,
):
    """

    Parameters
    ----------
    parm_dict: dict
        The dictionary in which the a parameter will be checked
    string_key :
    acceptable_data_types : list
    acceptable_data : list
    acceptable_range : list (length of 2)
    list_acceptable_data_types : list
    list_len : int
        If list_len is -1 than the list can be any length.
    default :
    Returns
    -------
    parm_passed : bool

    """

    import numpy as np

    if string_key in parm_dict:
        if (list in acceptable_data_types) or (np.array in acceptable_data_types):
            if (len(parm_dict[string_key]) != list_len) and (list_len != -1):
                print(
                    "######### ERROR:Parameter ",
                    string_key,
                    "must be a list of ",
                    list_acceptable_data_types,
                    " and length",
                    list_len,
                    ". Wrong length.",
                )
                return False
            else:
                list_len = len(parm_dict[string_key])
                for i in range(list_len):
                    type_check = False
                    for lt in list_acceptable_data_types:
                        if isinstance(parm_dict[string_key][i], lt):
                            type_check = True
                    if not (type_check):
                        print(
                            "######### ERROR:Parameter ",
                            string_key,
                            "must be a list of ",
                            list_acceptable_data_types,
                            " and length",
                            list_len,
                            ". Wrong type of",
                            type(parm_dict[string_key][i]),
                        )
                        return False

                    if acceptable_data is not None:
                        if not (parm_dict[string_key][i] in acceptable_data):
                            print(
                                "######### ERROR: Invalid",
                                string_key,
                                ". Can only be one of ",
                                acceptable_data,
                                ".",
                            )
                            return False

                    if acceptable_range is not None:
                        if (parm_dict[string_key][i] < acceptable_range[0]) or (
                            parm_dict[string_key][i] > acceptable_range[1]
                        ):
                            print(
                                "######### ERROR: Invalid",
                                string_key,
                                ". Must be within the range ",
                                acceptable_range,
                                ".",
                            )
                            return False
        elif dict in acceptable_data_types:
            parms_passed = True

            if default is None:
                print(
                    "######### ERROR:Dictionary parameters must have a default. Please report bug."
                )
                return False
            # print('is a dict',default)
            for default_element in default:
                if default_element in parm_dict[string_key]:
                    # print('1.*******')
                    # print(parm_dict[string_key], default_element, [type(default[default_element])], default[default_element])
                    if not (
                        check_parms(
                            parm_dict[string_key],
                            default_element,
                            [type(default[default_element])],
                            default=default[default_element],
                        )
                    ):
                        parms_passed = False
                    # print('2.*******')
                else:
                    # print('parm_dict',default_element,string_key)
                    parm_dict[string_key][default_element] = default[default_element]
        #                    print(
        #                        "Setting default",
        #                        string_key,
        #                        "['",
        #                        default_element,
        #                        "']",
        #                        " to ",
        #                        default[default_element],
        #                    )
        else:
            type_check = False
            for adt in acceptable_data_types:
                if isinstance(parm_dict[string_key], adt):
                    type_check = True
            if not (type_check):
                print(
                    "######### ERROR:Parameter ",
                    string_key,
                    "must be of type ",
                    acceptable_data_types,
                )
                return False

            if acceptable_data is not None:
                if not (parm_dict[string_key] in acceptable_data):
                    print(
                        "######### ERROR: Invalid",
                        string_key,
                        ". Can only be one of ",
                        acceptable_data,
                        ".",
                    )
                    return False

            if acceptable_range is not None:
                if (parm_dict[string_key] < acceptable_range[0]) or (
                    parm_dict[string_key] > acceptable_range[1]
                ):
                    print(
                        "######### ERROR: Invalid",
                        string_key,
                        ". Must be within the range ",
                        acceptable_range,
                        ".",
                    )
                    return False
    else:
        if default is not None:
            # print(parm_dict, string_key,  default)
            parm_dict[string_key] = default
            # print("Setting default", string_key, " to ", parm_dict[string_key])
        else:
            print("######### ERROR:Parameter ", string_key, "must be specified")
            return False

    return True


def _check_dataset(vis_dataset, data_variable_name):
    try:
        temp = vis_dataset[data_variable_name]
    except:
        print(
            "######### ERROR Data array ",
            data_variable_name,
            "can not be found in dataset.",
        )
        return False
    return True


def check_sel_parms(
    xds,
    sel_parms,
    new_or_modified_data_variables={},
    default_data_group_out=None,
    skip_data_group_in=False,
    skip_data_group_out=False,
):
    """

    Parameters
    ----------
    xds : xarray.core.dataset.Dataset
        Input vis.zarr multi dataset.
    sel_parms : dictionary
    Returns
    -------
    psf_dataset : xarray.core.dataset.Dataset
    """

    assert "data_groups" in xds.attrs, "No data_groups found in ms_xds."

    if "overwrite" not in sel_parms:
        sel_parms["overwrite"] = False
        overwrite = False
    else:
        overwrite = sel_parms["overwrite"]

    if not skip_data_group_in:
        if not ("data_group_in" in sel_parms):
            data_group_in_name = "base"
            sel_parms["data_group_in"] = {
                data_group_in_name: xds.attrs["data_groups"][data_group_in_name]
            }

        if isinstance(sel_parms["data_group_in"], str):
            data_group_in_name = sel_parms["data_group_in"]

            # print(data_group_in_name,xds.attrs["data_groups"])
            sel_parms["data_group_in"] = {
                sel_parms["data_group_in"]: xds.attrs["data_groups"][
                    sel_parms["data_group_in"]
                ]
            }
        else:
            data_group_in_name = list(sel_parms["data_group_in"].keys())[0]

        if data_group_in_name not in xds.attrs["data_groups"]:
            xds.attrs["data_groups"][data_group_in_name] = sel_parms["data_group_in"][
                data_group_in_name
            ]

        data_group_in = xds.attrs["data_groups"][data_group_in_name]
    else:
        data_group_in = None

    if default_data_group_out is not None:
        default_data_group_out_name = list(default_data_group_out.keys())[0]
        new_or_modified_data_variables = list(
            default_data_group_out[default_data_group_out_name].keys()
        )

        default_data_group_out[default_data_group_out_name] = {
            **sel_parms["data_group_in"][data_group_in_name],
            **default_data_group_out[default_data_group_out_name],
        }
    else:
        default_data_group_out = None

    if not skip_data_group_out:
        if not ("data_group_out_name" in sel_parms):
            sel_parms["data_group_out"] = default_data_group_out

        if isinstance(sel_parms["data_group_out"], str):
            sel_parms["data_group_out"] = {
                sel_parms["data_group_out"]: list(default_data_group_out.values)[0]
            }

        data_group_out_name = list(sel_parms["data_group_out"].keys())[0]

        if not overwrite:
            for nm_dv in new_or_modified_data_variables:
                assert (
                    sel_parms["data_group_out"][data_group_out_name][nm_dv] not in xds
                ), (
                    sel_parms["data_group_out"][data_group_out_name][nm_dv]
                    + " already present in xds. Set overwrite to True if data variable should be overwritten."
                )

            # assert data_group_out_name not in xds.attrs["data_groups"], "Data group " + data_group_out_name + " already in xds data_groups. Set overwrite to True if data group should be overwritten."

        xds.attrs["data_groups"][data_group_out_name] = sel_parms["data_group_out"][
            data_group_out_name
        ]

        data_group_out = xds.attrs["data_groups"][data_group_out_name]
    else:
        data_group_out = None

    return data_group_in, data_group_out


def _check_sub_sel_parms(sel_parms, select_defaults):
    parms_passed = True
    for sel in select_defaults:
        if not (
            check_parms(
                sel_parms,
                sel,
                [type(select_defaults[sel])],
                default=select_defaults[sel],
            )
        ):
            parms_passed = False
    return parms_passed


"""
def _check_sel_parms(sel_parms,select_defaults):
    parms_passed = True
    for sel_def in select_defaults:
        if isinstance(select_defaults[sel_def], dict):
            if sel_def in sel_parms:
                for sub_sel_def in select_defaults[sel_def]:
                        #print(sub_sel_def,select_defaults[sel_def])
                        #print(sel_parms[sel_def], sub_sel_def, select_defaults[sel_def][sub_sel_def])
                        if not(check_parms(sel_parms[sel_def], sub_sel_def, [str], default=select_defaults[sel_def][sub_sel_def])): parms_passed = False
            else:
                sel_parms[sel_def] = select_defaults[sel_def]
                print ('Setting default', string_key, ' to ', parm_dict[string_key])
        else:
            if not(check_parms(sel_parms, sel_def, [str], default=select_defaults[sel_def])): parms_passed = False
    return parms_passed
"""


def _check_existence_sel_parms(dataset, sel_parms):
    parms_passed = True
    for sel in sel_parms:
        if isinstance(sel_parms[sel], dict):
            if sel != "properties":
                _check_existence_sel_parms(dataset, sel_parms[sel])
        else:
            if (sel != "id") and (sel != "properties"):
                if not (_check_dataset(dataset, sel_parms[sel])):
                    parms_passed = False
    return parms_passed
